/*
 * Arithmetic coder
 * 
 * Copyright (c) 2018-2019 Fabrice Bellard
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <inttypes.h>
#include <assert.h>
#include <time.h>
#include <getopt.h>

#include "cutils.h"
#include "arith.h"

#define SYM_COUNT 256
#define RANGE_MIN SYM_COUNT

void put_bit_init(PutBitState *s,
                  PutBitWriteFunc *write_func, void *opaque)
{
    memset(s, 0, sizeof(*s));
    s->write_func = write_func;
    s->opaque = opaque;
    s->low = 0;
    s->range = SYM_COUNT * SYM_COUNT;
}

/* XXX: do progressive flush */
static void put_bit_resize(PutBitState *s, int len)
{
    int new_size;
    if (len > s->buf_size) {
        new_size = max_int(len, s->buf_size * 3 / 2);
        s->buf = realloc(s->buf, new_size);
        s->buf_size = new_size;
    }
}

static no_inline void put_bit_flush(PutBitState *s)
{
    if (s->idx > 0) {
        s->write_func(s->opaque, s->buf, s->idx);
        s->idx = 0;
    }
}

static void put_bit_renorm(PutBitState *s)
{
    int v, carry, i;
    while (s->range < RANGE_MIN) {
        v = s->low / SYM_COUNT;
#ifdef DUMP_PUT_BIT
        printf("  out=%d\n", v % SYM_COUNT);
#endif
        
        put_bit_resize(s, s->idx + 1);
        s->buf[s->idx++] = v % SYM_COUNT;
        carry = v / SYM_COUNT;
        if (carry) {
#ifdef DUMP_PUT_BIT
            printf("  carry\n");
#endif
            assert(carry <= 1);
            i = s->idx - 2;
            while (carry && i >= 0) {
                v = s->buf[i] + carry;
                s->buf[i] = v % SYM_COUNT;
                carry = v >= SYM_COUNT;
                i--;
            }
            assert(carry == 0);
        }
        s->low = (s->low % SYM_COUNT) * SYM_COUNT;
        s->range *= SYM_COUNT;
    }
}

/* 0 < prob0 < PROB_UNIT */
void put_bit(PutBitState *s, int prob0, int bit)
{
    int range0;

    assert(s->range >= RANGE_MIN);
    range0 = (s->range * prob0) >> 8;
    assert(range0 > 0);
    assert(range0 < s->range);
#ifdef DUMP_PUT_BIT
    {
        static int count;
        printf("%d: c=%d range=%d b=%d range0=%d low=%d\n",
               count++, c, s->range, bit, range0, s->low);
    }
#endif
    if (!bit) {
        s->range = range0;
    } else {
        s->low += range0;
        s->range -= range0;
    }
    
    put_bit_renorm(s);
}

void put_bit_raw(PutBitState *s, int bit)
{
    int range0;
    
    assert(s->range >= RANGE_MIN);
    range0 = s->range >> 1;
    if (!bit) {
        s->range = range0;
    } else {
        s->low += range0;
        s->range -= range0;
    }
    
    put_bit_renorm(s);
}

void put_bit_end(PutBitState *s)
{
    s->range = RANGE_MIN - 1;
    put_bit_renorm(s);
    s->range = RANGE_MIN - 1;
    put_bit_renorm(s);
    put_bit_flush(s);
    free(s->buf);
}

double put_bit_get_bit_count(PutBitState *s)
{
    return s->idx * 8.0 + log2((SYM_COUNT * SYM_COUNT) / (double)s->range);
}

/****************************************/

static void refill(GetBitState *s)
{
    s->range *= SYM_COUNT;
    s->low = s->low * SYM_COUNT;
    if (s->idx >= s->buf_len) {
        if (!s->read_func)
            return; /* pad with zeros */
        s->buf_len = s->read_func(s->opaque, s->buf, s->buf_size);
        s->idx = 0;
    }
#ifdef DUMP_GET_BIT
    printf("  in=%d\n", s->buf[s->idx]);
#endif
    s->low += s->buf[s->idx++];
}

void get_bit_init(GetBitState *s, uint8_t *buf, size_t buf_size,
                  GetBitReadFunc *read_func, void *opaque)
{
    s->buf_size = buf_size;
    s->buf = buf;
    s->read_func = read_func;
    s->opaque = opaque;
    if (read_func) {
        s->buf_len = 0;
    } else {
        /* prefilled buffer */
        s->buf_len = s->buf_size;
    }
    s->range = 1;
    s->low = 0;
    s->idx = 0;
    refill(s);
    refill(s);
}

/* 0 < prob0 < PROB_UNIT */
int get_bit(GetBitState *s, int prob0)
{
    int b, range0;

    assert(s->range >= RANGE_MIN);
    range0 = (s->range * prob0) >> 8;
    assert(range0 > 0);
    assert(range0 < s->range);
    b = s->low >= range0;
#ifdef DUMP_GET_BIT
    {
        static int count;
        printf("%d: c=%d range=%d b=%d range0=%d low=%d\n", count++, c, s->range, b, range0, s->low);
    }
#endif
    if (b) {
        s->low -= range0;
        s->range -= range0;
    } else {
        s->range = range0;
    }
    if (s->range < RANGE_MIN)
        refill(s);
    return b;
}

/* no context */
int get_bit_raw(GetBitState *s)
{
    int b, range0;
    range0 = s->range >> 1;
    b = s->low >= range0;
    if (b) {
        s->low -= range0;
        s->range -= range0;
    } else {
        s->range = range0;
    }
    if (s->range < RANGE_MIN)
        refill(s);
    return b;
}
