/*
 * Compression utilities
 * 
 * Copyright (c) 2018-2019 Fabrice Bellard
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <inttypes.h>
#include <assert.h>
#include <time.h>
#include <getopt.h>
#include <stdarg.h>
#include <sys/time.h>
#include <sys/stat.h>
#ifdef _WIN32
#include <direct.h>
#endif

#include "cutils.h"
#include "libnc.h"
#include "cp_utils.h"

int64_t get_time_ms(void)
{
#ifdef _WIN32
    struct timeval tv;
    gettimeofday(&tv, NULL);
    return (int64_t)tv.tv_sec * 1000 + (tv.tv_usec / 1000U);
#else
    struct timespec ts;
    clock_gettime(CLOCK_MONOTONIC, &ts);
    return (int64_t)ts.tv_sec * 1000 + (ts.tv_nsec / 1000000U);
#endif
}

void fput_u8(FILE *f, uint8_t v)
{
    fputc(v, f);
}

int fget_u8(FILE *f, uint8_t *pv)
{
    int c;
    c = fgetc(f);
    if (c < 0)
        return -1;
    *pv = c;
    return 0;
}

void fput_be16(FILE *f, uint16_t v)
{
    fputc(v >> 8, f);
    fputc(v >> 0, f);
}

int fget_be16(FILE *f, uint16_t *pv)
{
    uint8_t buf[2];
    if (fread(buf, 1, sizeof(buf), f) != sizeof(buf))
        return -1;
    *pv = (buf[0] << 8) |
        (buf[1] << 0);
    return 0;
}

void fput_be32(FILE *f, uint32_t v)
{
    fputc(v >> 24, f);
    fputc(v >> 16, f);
    fputc(v >> 8, f);
    fputc(v >> 0, f);
}

int fget_be32(FILE *f, uint32_t *pv)
{
    uint8_t buf[4];
    if (fread(buf, 1, sizeof(buf), f) != sizeof(buf))
        return -1;
    *pv = (buf[0] << 24) |
        (buf[1] << 16) |
        (buf[2] << 8) |
        (buf[3] << 0);
    return 0;
}

void fput_sgd_opt(FILE *f, const SGDOptParams *p)
{
    fput_u8(f, p->algo);
    switch(p->algo) {
    case SGD_OPT_BASIC:
        break;
    case SGD_OPT_ADAM:
        fput_f32(f, p->u.adam.beta1);
        fput_f32(f, p->u.adam.beta2);
        fput_f32(f, p->u.adam.eps);
        break;
    default:
        abort();
    }

    fput_f32(f, p->learning_rate);
}
                  
int fget_sgd_opt(FILE *f, SGDOptParams *p)
{
    uint8_t v8;

    if (fget_u8(f, &v8))
        return -1;
    p->algo = v8;
    switch(p->algo) {
    case SGD_OPT_BASIC:
        break;
    case SGD_OPT_ADAM:
        if (fget_f32(f, &p->u.adam.beta1))
            return -1;
        if (fget_f32(f, &p->u.adam.beta2))
            return -1;
        if (fget_f32(f, &p->u.adam.eps))
            return -1;
        break;
    default:
        return -1;
    }
    if (fget_f32(f, &p->learning_rate))
        return -1;
    return 0;
}

typedef union {
    float f;
    uint32_t u32;
} f32;

void fput_f32(FILE *f, float v)
{
    f32 u;
    u.f = v;
    fput_be32(f, u.u32);
}

int fget_f32(FILE *f, float *pv)
{
    f32 u;
    if (fget_be32(f, &u.u32))
        return -1;
    *pv = u.f;
    return 0;
}

void write_sym(PutBitState *pb, const float *prob_table, int n_symb, int sym)
{
    int start, range, prob0, bit, range0;
    float p, p0;
    
    start = 0;
    range = n_symb;
    p = 1.0; /* invariant: p=sum(prob_table[start...start + range]) */
    while (range > 1) {
        range0 = range >> 1;
        p0 = vec_sum(prob_table + start, range0);
        prob0 = lrintf(p0 * PROB_UNIT / p);
        prob0 = clamp_int(prob0, 1, PROB_UNIT - 1);
        bit = sym >= (start + range0);
        put_bit(pb, prob0, bit);
        if (bit) {
            start += range0;
            range = range - range0;
            p = p - p0;
        } else {
            p = p0;
            range = range0;
        }
    }
}

int read_sym(GetBitState *gb, const float *prob_table, int n_symb)
{
    int start, range, prob0, bit, range0;
    float p, p0;
    
    start = 0;
    range = n_symb;
    p = 1.0; /* invariant: p=sum(prob_table[start...start + range]) */
    while (range > 1) {
        range0 = range >> 1;
        p0 = vec_sum(prob_table + start, range0);
        prob0 = lrintf(p0 * PROB_UNIT / p);
        prob0 = clamp_int(prob0, 1, PROB_UNIT - 1);
        bit = get_bit(gb, prob0);
        if (bit) {
            start += range0;
            range = range - range0;
            p = p - p0;
        } else {
            p = p0;
            range = range0;
        }
    }
    return start;
}

static void create_dir(Stats *st, const char *debug_dir,
                       const char *prefix, const char *filename)
{
    const char *p;
    char name1[1024];
    struct tm *tm;
    time_t ti;
    
    p = strrchr(filename, '/');
    if (!p)
        p = filename;
    else
        p++;
    snprintf(name1, sizeof(name1), "%s/%s-%s", debug_dir, prefix, p);
#ifdef _WIN32
    _mkdir(name1);
#else
    mkdir(name1, 0777);
#endif
    
    ti = time(NULL);
    tm = gmtime(&ti);
    snprintf(st->debug_dir, sizeof(st->debug_dir), "%s/%04u%02u%02u-%02u%02u%02u",
             name1,
             tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday,
             tm->tm_hour, tm->tm_min, tm->tm_sec);
#ifdef _WIN32
    _mkdir(st->debug_dir);
#else
    mkdir(st->debug_dir, 0777);
#endif
}

void stats_init(Stats *st, const char *debug_dir, const char *prog_name,
                const char *in_filename, int stats_avg_count)
{
    char filename[1024];
    
    memset(st, 0, sizeof(*st));
    if (debug_dir) {
        st->debug_output = TRUE;
        create_dir(st, debug_dir, prog_name, in_filename);
        printf("[Outputing logs to '%s']\n", st->debug_dir);
        snprintf(filename, sizeof(filename), "%s/plot.txt", st->debug_dir);
        st->fp = fopen(filename, "wb");
        if (!st->fp) {
            fprintf(stderr, "could not create '%s'\n", filename);
            exit(1);
        }
    }
    st->start_time = st->last_time = get_time_ms();
    st->stats_avg_count = stats_avg_count;
    st->bps_buf = nc_mallocz(sizeof(st->bps_buf[0]) * st->stats_avg_count);
}

void stats_dump(Stats *st, NCModel *m, double n_output_bits)
{
    int64_t ti;
    double n_output_bits1, bps_avg, bps, bps_tot;
    int n_input_bytes1, i;
    
    n_input_bytes1 = st->n_input_bytes - st->last_n_input_bytes;
    n_output_bits1 = n_output_bits - st->last_n_output_bits;
    ti = get_time_ms() - st->last_time;
    if (ti <= 0)
        ti = 1;
    
    bps = n_output_bits1 / n_input_bytes1;

    /* average bps over STATS_AVG_COUNT */
    st->bps_buf[st->bps_buf_idx] = bps;
    if (++st->bps_buf_idx == st->stats_avg_count)
        st->bps_buf_idx = 0;
    if (st->bps_buf_len < st->stats_avg_count)
        st->bps_buf_len++;
    bps_avg = 0;
    for(i = 0; i < st->bps_buf_len; i++)
        bps_avg += st->bps_buf[i];
    bps_avg /= st->bps_buf_len;

    bps_tot = n_output_bits / st->n_input_bytes;
    
    if (!st->header_output) {
        st->header_output = TRUE;
        printf("%10s %6s %6s %6s %6s %8s\n",
               "SIZE", "locBPS", "BPS/5M", "BPS", "kS/s", "LR");
        if (st->fp) {
            fprintf(st->fp, "# %10s %6s %6s %6s %6s %8s\n",
                    "SIZE", "locBPS", "BPS/5M", "BPS", "kS/s", "LR");
        }
    }
    printf("%10" PRIu64 " %6.3f %6.3f %6.3f %6.2f %8.2e\r",
           st->n_input_bytes,
           bps,
           bps_avg,
           bps_tot,
           (double)n_input_bytes1 / ti,
           st->learning_rate);
    fflush(stdout);
    if (st->fp) {
        fprintf(st->fp, "%10" PRIu64 " %6.3f %6.3f %6.3f %6.2f %8.2e\n",
                st->n_input_bytes,
                bps,
                bps_avg,
                bps_tot,
                (double)n_input_bytes1 / ti,
                st->learning_rate);
        fflush(st->fp);
    }
    st->last_time = get_time_ms();
    st->last_n_output_bits = n_output_bits;
    st->last_n_input_bytes = st->n_input_bytes;
}

void stats_end(Stats *st, NCModel *m, int64_t n_output_bytes)
{
    int64_t ti;
    ti = get_time_ms() - st->start_time;
    printf("\nbps=%0.3f (%" PRId64 " bytes), %0.2f kS/s\n",
           (double)n_output_bytes * 8 / st->n_input_bytes,
           n_output_bytes,
           (double)st->n_input_bytes / ti);
    if (st->fp) {
        fprintf(st->fp, "# bps=%0.3f (%" PRId64 " bytes), %0.2f kS/s\n",
                (double)n_output_bytes * 8 / st->n_input_bytes,
                n_output_bytes,
                (double)st->n_input_bytes / ti);
        fclose(st->fp);
    }
    nc_free(st->bps_buf);
    if (st->debug_output) { 
        char filename[1024];
        snprintf(filename, sizeof(filename), "%s/coefs.bin", st->debug_dir);
        nc_save_coefs(m, filename);
    }
}

/* we print at least 3 significant digits with at most 5 chars, except
   if larger than 9999T. The value is rounded to zero. */
char *get_si_prefix(char *buf, int buf_size, uint64_t val)
{
    static const char suffixes[4] = "kMGT";
    uint64_t base;
    int i;

    if (val <= 999) {
        snprintf(buf, buf_size, "%" PRId64, val);
    } else {
        base = 1000;
        for(i=0;i<4;i++) {
            /* Note: we round to 0 */
            if (val < base * 10) {
                snprintf(buf, buf_size, "%0.2f%c", 
                         floor((val * 100.0) / base) / 100.0,
                         suffixes[i]);
                break;
            } else if (val < base * 100) {
                snprintf(buf, buf_size, "%0.1f%c", 
                         floor((val * 10.0) / base) / 10.0,
                         suffixes[i]);
                break;
            } else if (val < base * 1000 || (i == 3)) {
                snprintf(buf, buf_size,
                         "%" PRId64 "%c", 
                         val / base,
                         suffixes[i]);
                break;
            }
            base = base * 1000;
        }
    }
    return buf;
}
