/*
 * Lossless compression with LSTM
 * 
 * Copyright (c) 2018-2019 Fabrice Bellard
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <inttypes.h>
#include <assert.h>
#include <time.h>
#include <getopt.h>
#include <stdarg.h>
#include <sys/time.h>

#include "cutils.h"
#include "arith.h"
#include "libnc.h"
#include "cp_utils.h"

/************************************************/
/* LSTM cell */

typedef enum {
    LSTM_TYPE_NORMAL,
    LSTM_TYPE_CLAMPED,
    LSTM_TYPE_TIED,
    LSTM_TYPE_GRU,
} LSTMTypeEnum;

typedef struct {
    int n_inputs;
    int n_sparse_inputs;
    int n_cells;
    int n_cells2; /* >= n_cells, > n_cells means to use projection */
    BOOL use_layer_norm;
    LSTMTypeEnum lstm_type;
} LSTMLayerParams;

#define LSTM_MAT_COUNT_MAX 4

#define LSTM_FORGET_GATE 0
#define LSTM_INPUT_NODE  1
#define LSTM_OUTPUT_GATE 2
#define LSTM_INPUT_GATE  3

#define GRU_UPDATE_GATE 0
#define GRU_RESET_GATE  1
#define GRU_OUTPUT_NODE 2

typedef struct {
    BOOL use_layer_norm;
    int mat_count;
    NCSignal *u;
    NCSignal *w;
    NCSignal *ws;
    NCSignal *b[LSTM_MAT_COUNT_MAX];
    NCSignal *g[LSTM_MAT_COUNT_MAX]; /* only for layer norm */
    NCSignal *p; /* only used for projection */
} LSTMCell;

static LSTMCell *lstm_cell_init(NCModel *m, LSTMLayerParams *p, int layer_idx)
{
    LSTMCell *s;
    int j;
    s = nc_mallocz(sizeof(*s));
    s->use_layer_norm = p->use_layer_norm;
    assert(p->n_cells2 >= p->n_cells);
    if (p->lstm_type == LSTM_TYPE_TIED || p->lstm_type == LSTM_TYPE_GRU) {
        s->mat_count = 3;
    } else {
        s->mat_count = 4;
    }
    s->u = nc_new_var_2d(m, NC_TYPE_F32, p->n_cells2 * s->mat_count,
                         p->n_cells);
    nc_signal_set_name(s->u, "u%d", layer_idx);
    if (p->n_inputs != 0) {
        s->w = nc_new_var_2d(m, NC_TYPE_F32, p->n_cells2 * s->mat_count,
                             p->n_inputs);
        nc_signal_set_name(s->w, "w%d", layer_idx);
    }
    if (p->n_sparse_inputs != 0) {
        s->ws = nc_new_var_2d(m, NC_TYPE_F32, p->n_cells2 * s->mat_count,
                              p->n_sparse_inputs);
        nc_signal_set_name(s->ws, "ws%d", layer_idx);
    }
    for(j = 0; j < s->mat_count; j++) {
        s->b[j] = nc_new_var_1d(m, NC_TYPE_F32, p->n_cells2);
        nc_signal_set_name(s->b[j], "b%d_%d", layer_idx, j);
        if (s->use_layer_norm) {
            s->g[j] = nc_new_var_1d(m, NC_TYPE_F32, p->n_cells2);
            nc_signal_set_name(s->g[j], "g%d_%d", layer_idx, j);
        }
    }
    if (p->n_cells2 != p->n_cells) {
        s->p = nc_new_var_2d(m, NC_TYPE_F32, p->n_cells, p->n_cells2);
        nc_signal_set_name(s->p, "p%d", layer_idx);
    }
    return s;
}

/************************************************/
/* LSTM model */

typedef struct {
    BOOL is_decoder;
    int n_layers;
    int n_cells;
    int n_cells2;
    int n_embed_out;
    int n_symbols;
    int n_states; 
    int n_streams;
    LSTMTypeEnum lstm_type;
    BOOL use_layer_norm;
    BOOL full_connect;
    SGDOptParams sgd_opt;
    uint32_t seed;
    int block_len;
} LSTMParams;

typedef struct {
    RNDState rnd_state;
    int n_layers;
    int n_inputs;
    int n_outputs;
    int n_cells;
    int n_streams;
    int n_states;
    LSTMTypeEnum lstm_type;
    NCModel *model;
    NCSignal **input, *expected_output, *output;
    NCSignal **c_in, **h_in;
    NCSignal **c_out, **h_out;
    float *prob_output;
    size_t prob_output_stride;
    NCCompileResult func;
    NCSGDOptState *sgd_opt;
} LSTM;

static void lstm_reset(LSTM *s);

static int nb_threads = 1;

static void var_init_rnd(NCSignal *x, float range, RNDState *rnd_state)
{
    NCSignalData dbuf, *d;
    d = nc_signal_get_data(&dbuf, x);
    rnd_unif_mat(d->data, d->stride,
                 d->n_strides, d->dims[0],
                 0.0f, range,
                 rnd_state);
}

static NCSignal *concat_add(NCSignal **tab, int n_in, int n_out)
{
    NCSignal **tab1, *out;
    int q, r, i, n, j, k;
    assert(n_out <= n_in);
    tab1 = nc_mallocz(sizeof(tab1[0]) * n_out);
    q = n_in / n_out;
    r = n_in % n_out;
    /* if r != 0, we add less high order inputs */
    k = 0;
    for(i = 0; i < n_out; i++) {
        n = q + (i < r);
        tab1[i] = tab[k++];
        for(j = 1; j < n; j++) {
            tab1[i] = nc_add(tab1[i], tab[k++]);
        }
    }
    assert(k == n_in);
    out = nc_vconcat(tab1, n_out);
    nc_free(tab1);
    return out;
}

static LSTM *lstm_init(const LSTMParams *p)
{
    LSTM *s;
    int layer_idx;
    NCModel *m;
    NCSignal *t0, *fc_b, *fc_w, *c, *h, **layer_output, **tab_tmp;
    NCSignal **tab_tmp1;
    LSTMCell **lstm_layers;
    int state_idx, i, j;
    
    s = nc_mallocz(sizeof(*s));
    rnd_init(&s->rnd_state, p->seed);
    s->n_states = p->n_states;
    s->n_streams = p->n_streams;
    s->n_layers = p->n_layers;
    s->n_inputs = p->n_symbols;
    s->n_outputs = p->n_symbols;
    s->n_cells = p->n_cells;
    s->lstm_type = p->lstm_type;
    
    m = nc_model_init(nb_threads);
    s->model = m;

    nc_start_func(m, "eval");

    lstm_layers = nc_mallocz(sizeof(lstm_layers[0]) * s->n_layers);
    for(layer_idx = 0; layer_idx < s->n_layers; layer_idx++) {
        LSTMLayerParams lp_s, *lp = &lp_s;
        memset(lp, 0, sizeof(*lp));
        if (p->full_connect) {
            /* input from all previous layers */
            lp->n_inputs = s->n_cells * layer_idx;
        } else {
            /* input from the previous layer */
            lp->n_inputs = layer_idx != 0 ? s->n_cells : 0;
        }
        lp->n_sparse_inputs = s->n_inputs;
        lp->n_cells = s->n_cells;
        lp->n_cells2 = p->n_cells2;
        lp->use_layer_norm = p->use_layer_norm;
        lp->lstm_type = p->lstm_type;
        lstm_layers[layer_idx] = lstm_cell_init(m, lp, layer_idx);
    }
    
    fc_b = nc_new_var_1d(m, NC_TYPE_F32, s->n_outputs);
    nc_signal_set_name(fc_b, "fc_b");
    fc_w = nc_new_var_2d(m, NC_TYPE_F32, s->n_outputs, s->n_cells * p->n_embed_out);
    nc_signal_set_name(fc_w, "fc_w");
    s->input = nc_mallocz(sizeof(s->input[0]) * s->n_states);
    for(state_idx = 0; state_idx < s->n_states; state_idx++) {
        s->input[state_idx] = nc_new_local_1d(m, NC_TYPE_I32, s->n_streams);
        nc_signal_set_global(s->input[state_idx]);
        nc_signal_set_name(s->input[state_idx], "input%d", state_idx);
    }
    s->expected_output =
        nc_signal_set_name(nc_new_local_2d(m, NC_TYPE_I32, s->n_streams, s->n_states),
                                   "eoutput");
    nc_signal_set_global(s->expected_output);

    layer_output = nc_mallocz(sizeof(layer_output[0]) * s->n_states * s->n_layers);
    if (s->lstm_type != LSTM_TYPE_GRU) {
        s->c_in = nc_mallocz(sizeof(s->c_in[0]) * s->n_layers);
        s->c_out = nc_mallocz(sizeof(s->c_out[0]) * s->n_layers);
        for(i = 0; i < s->n_layers; i++) {
            s->c_in[i] = nc_new_local_2d(m, NC_TYPE_F32, p->n_cells2, s->n_streams);
            nc_signal_set_global(s->c_in[i]);
            nc_signal_set_name(s->c_in[i], "c_in%d", i);
        }
    }

    s->h_in = nc_mallocz(sizeof(s->h_in[0]) * s->n_layers);
    s->h_out = nc_mallocz(sizeof(s->h_out[0]) * s->n_layers);
    for(i = 0; i < s->n_layers; i++) {
        s->h_in[i] = nc_new_local_2d(m, NC_TYPE_F32, s->n_cells, s->n_streams);
        nc_signal_set_global(s->h_in[i]);
        nc_signal_set_name(s->h_in[i], "h_in%d", i);
    }

    tab_tmp = nc_mallocz(sizeof(tab_tmp[0]) *
                         max_int(LSTM_MAT_COUNT_MAX * s->n_states,
                                 s->n_layers));
    tab_tmp1 = nc_mallocz(sizeof(tab_tmp1[0]) * s->n_states);

    for(layer_idx = 0; layer_idx < s->n_layers; layer_idx++) {
        LSTMCell *lc = lstm_layers[layer_idx];

        /* handle the inputs first for cache locality */
        if (layer_idx != 0) {
            for(state_idx = 0; state_idx < s->n_states; state_idx++) {
                if (!p->full_connect || layer_idx == 1) {
                    t0 = layer_output[state_idx * s->n_layers + layer_idx - 1];
                } else {
                    for(i = 0; i < layer_idx; i++)
                        tab_tmp[i] = layer_output[state_idx * s->n_layers + i];
                    t0 = nc_vconcat(tab_tmp, layer_idx);
                }
                tab_tmp1[state_idx] = t0;
            }
            t0 = nc_hconcat(tab_tmp1, s->n_states);
            t0 = nc_matmul(lc->w, t0);
            nc_hsplit(tab_tmp, t0, s->n_states, NULL);
        }
        
        h = s->h_in[layer_idx];
        if (s->lstm_type == LSTM_TYPE_GRU)
            c = NULL;
        else
            c = s->c_in[layer_idx];
        for(state_idx = 0; state_idx < s->n_states; state_idx++) {
            NCSignal *tab[LSTM_MAT_COUNT_MAX];

            t0 = nc_get_col(lc->ws, s->input[state_idx]);
            /* XXX: GRU is no longer supported */
            t0 = nc_add(t0, nc_matmul(lc->u, h));
            if (layer_idx != 0)
                t0 = nc_add(t0, tab_tmp[state_idx]);
            nc_vsplit(tab, t0, lc->mat_count, NULL);
            
            for(j = 0; j < lc->mat_count; j++) {
                t0 = tab[j];
                if (p->use_layer_norm) {
                    t0 = nc_layer_norm(t0);
                    t0 = nc_mul(t0, nc_dup(lc->g[j], s->n_streams));
                }
                t0 = nc_add(t0, lc->b[j]);
                tab[j] = t0;
            }

            if (s->lstm_type == LSTM_TYPE_GRU) {
                h = nc_interp(h, nc_tanh(tab[GRU_OUTPUT_NODE]),
                              nc_sigmoid(tab[GRU_UPDATE_GATE]));
            } else {
                NCSignal *fg, *ig, *og, *in;
                fg = nc_sigmoid(tab[LSTM_FORGET_GATE]);
                og = nc_sigmoid(tab[LSTM_OUTPUT_GATE]);
                in = nc_tanh(tab[LSTM_INPUT_NODE]);
                switch(s->lstm_type) {
                case LSTM_TYPE_NORMAL:
                    ig = nc_sigmoid(tab[LSTM_INPUT_GATE]);
                    c = nc_add(nc_mul(c, fg), nc_mul(in, ig));
                    h = nc_mul(og, nc_tanh(c));
                    break;
                case LSTM_TYPE_CLAMPED:
                    ig = nc_sigmoid(tab[LSTM_INPUT_GATE]);
                    c = nc_lstm_clamped(c, in, fg, ig);
                    h = nc_mul(og, c);
                    break;
                case LSTM_TYPE_TIED:
                    c = nc_interp(c, in, fg);
                    h = nc_mul(og, c);
                    break;
                default:
                    abort();
                }
            }
            /* optional projection */
            if (lc->p)
                h = nc_matmul(lc->p, h);

            nc_signal_set_name(h, "h%d_%d", layer_idx, state_idx);
            layer_output[state_idx * s->n_layers + layer_idx] = h;
        }
        if (c) {
            nc_signal_set_global(c);
            s->c_out[layer_idx] = c;
            nc_signal_set_name(c, "c_out%d", layer_idx);
        }
        nc_signal_set_global(h);
        s->h_out[layer_idx] = h;
        nc_signal_set_name(h, "h_out%d", layer_idx);
    }
    for(state_idx = 0; state_idx < s->n_states; state_idx++) {
        tab_tmp1[state_idx] = concat_add(layer_output + state_idx * s->n_layers,
                                         s->n_layers, p->n_embed_out);
    }
    t0 = nc_hconcat(tab_tmp1, s->n_states);
    t0 = nc_matmul(fc_w, t0);
    t0 = nc_add(t0, fc_b);
    t0 = nc_reshape_3d(t0, s->n_outputs, s->n_streams, s->n_states);
    t0 = nc_log_soft_max(t0, s->expected_output);
    nc_signal_set_name(t0, "output");
    nc_signal_set_global(t0);
    s->output = t0;

    nc_free(layer_output);
    nc_free(tab_tmp);
    nc_free(tab_tmp1);
    {
        NCCompileParams sp_s, *sp = &sp_s;
        memset(sp, 0, sizeof(*sp));
        sp->grad = TRUE;
        if (p->is_decoder) {
            sp->partial_eval = TRUE;
            sp->output = s->output;
            sp->output_axis = 2;
        }
        nc_compile(m, &s->func, sp);
    }
    
    /* init the variables */
    {
        NCSignal *w;
        NCSignalData dbuf, *d;
        size_t j;
        for(layer_idx = 0; layer_idx < s->n_layers; layer_idx++) {
            LSTMCell *ls = lstm_layers[layer_idx];
            w = ls->u;
            if (w) {
                var_init_rnd(w, 1.0f / sqrtf(s->n_cells),
                             &s->rnd_state);
            }
            w = ls->w;
            if (w) {
                d = nc_signal_get_data(&dbuf, w);
                var_init_rnd(w, 1.0f / sqrtf(d->dims[1]),
                             &s->rnd_state);
            }
            w = ls->ws;
            var_init_rnd(w, 0.75f,
                         &s->rnd_state);
            for(i = 0; i < ls->mat_count; i++) {
                w = ls->g[i];
                if (w) {
                    NCSignalData dbuf, *d;
                    d = nc_signal_get_data(&dbuf, w);
                    for(j = 0; j < s->n_cells; j++)
                        ((float *)d->data)[j] = 1.0f;
                }
                w = ls->p;
                if (w) {
                    var_init_rnd(w, sqrtf(6.0f / (s->n_cells + p->n_cells2)),
                                 &s->rnd_state);
                }
            }
        }
        
        w = fc_w;
        var_init_rnd(w, sqrtf(12.0f / (s->n_cells * s->n_layers)),
                     &s->rnd_state);
    }

    /* apply the SGD optimizer to all the parameters */
    s->sgd_opt = nc_sgd_opt_init(m, &p->sgd_opt);

    nc_sgd_opt_set_all(m, s->sgd_opt);

    for(layer_idx = 0; layer_idx < s->n_layers; layer_idx++) {
        LSTMCell *ls = lstm_layers[layer_idx];
        nc_free(ls);
    }
    nc_free(lstm_layers);

    {
        NCSignalData dbuf, *d;
        d = nc_signal_get_data(&dbuf, s->output);
        s->prob_output = d->data;
        s->prob_output_stride = d->stride;
    }
    lstm_reset(s);
    return s;
}

void lstm_end(LSTM *s)
{
    nc_free(s->input);
    nc_free(s->h_in);
    nc_free(s->c_in);
    nc_free(s->h_out);
    nc_free(s->c_out);
    nc_free(s->func.partial_eval_func);
    nc_model_end(s->model);
    nc_free(s);
}

static void lstm_reset(LSTM *s)
{
    int stream_idx;
    float a;
    int i;
    for(stream_idx = 0; stream_idx < s->n_streams; stream_idx++) {
        float *output = s->prob_output + stream_idx * s->prob_output_stride;
        /* for the first char */
        a = 1.0f / s->n_outputs;
        for(i = 0; i < s->n_outputs; i++)
            output[i] = a;
    }
    for(i = 0; i < s->n_layers; i++) {
        nc_signal_set_f32(s->h_in[i], 0);
        if (s->c_in) {
            nc_signal_set_f32(s->c_in[i], 0);
        }
    }
}

static void lstm_eval(LSTM *s, int output_index,
                      const int *input_syms)
{
    int stream_idx, state_idx1, n_states, state_start;
    NCSignalData xbuf, *x;
    int32_t *idata;
    
    prof_start(PROF_EVAL);
    
    if (output_index < 0) {
        n_states = s->n_states;
        state_start = 0;
    } else {
        n_states = 1;
        state_start = output_index;
    }
        
    for(state_idx1 = 0; state_idx1 < n_states; state_idx1++) {
        x = nc_signal_get_data(&xbuf, s->input[state_start + state_idx1]);
        idata = (int32_t *)x->data;
        for(stream_idx = 0; stream_idx < s->n_streams; stream_idx++) {
            idata[stream_idx] = input_syms[state_idx1 * s->n_streams + stream_idx];
        }
    }
    
    if (output_index < 0) {
        nc_eval(s->func.eval_func);
    } else {
        nc_eval(s->func.partial_eval_func[output_index]);
    }
    
    prof_end(PROF_EVAL);
}

/* evaluate the error gradient (stored in x_update fields) */
static void lstm_eval_gradient(LSTM *s, const int *output_syms)
{
    int state_idx, stream_idx;
    NCSignalData xbuf, *x;
    int32_t *idata;
    
    prof_start(PROF_GRAD);

    x = nc_signal_get_data(&xbuf, s->expected_output);
    idata = (int32_t *)x->data;
    for(state_idx = 0; state_idx < s->n_states; state_idx++) {
        for(stream_idx = 0; stream_idx < s->n_streams; stream_idx++) {
            idata[state_idx * x->stride + stream_idx] =
                output_syms[state_idx * s->n_streams + stream_idx];
        }
    }
    nc_eval(s->func.grad_func);
    prof_end(PROF_GRAD);
}

/* update with coefficients */
static void lstm_update(LSTM *s)
{
    int layer_idx;

    prof_start(PROF_UPDATE);
    nc_update(s->model);

    /* copy the output state to the input state for the next evaluation */
    for(layer_idx = 0; layer_idx < s->n_layers; layer_idx++) {
        nc_signal_copy(s->h_in[layer_idx], s->h_out[layer_idx]);
    }
    if (s->c_in) {
        for(layer_idx = 0; layer_idx < s->n_layers; layer_idx++) {
            nc_signal_copy(s->c_in[layer_idx], s->c_out[layer_idx]);
        }
    }

    prof_end(PROF_UPDATE);
}

static double lstm_eval_error(LSTM *s, const int *input_syms,
                              const int *output_syms)
{
    double E;
    int state_idx, layer_idx, stream_idx, c;
    RNDState rnd_state;
    NCSignal *x;
    
    /* random initial state */
    rnd_init(&rnd_state, 3);
    for(layer_idx = 0; layer_idx < s->n_layers; layer_idx++) {
        x = s->h_in[layer_idx];
        var_init_rnd(x, 0.2, &rnd_state);
        x = s->c_in[layer_idx];
        var_init_rnd(x, 0.2, &rnd_state);
    }

    E = 0;
    for(state_idx = 0; state_idx < s->n_states; state_idx++) {
        lstm_eval(s, state_idx,
                  input_syms + state_idx * s->n_streams);
        for(stream_idx = 0; stream_idx < s->n_streams; stream_idx++) {
            c = output_syms[state_idx * s->n_streams + stream_idx];
            E -= log(s->prob_output[(state_idx * s->n_streams + stream_idx) *
                                    s->prob_output_stride + c]);
        }
    }
    return E;
}

void gradient_test(const char *coef_name, float eps)
{
    LSTMParams p_s, *p = &p_s;
    int n_syms, *input_syms, *output_syms, it, coefs_x, coefs_y, i;
    LSTM *s;
    double E0, E1;
    float grad_est, saved_coef, grad, *pcoef, *pgrad;
    float ratio;
    NCSignal *coef1;
    RNDState rnd_state;
    NCSignalData coef_buf, *coef;

    n_syms = 5;

    memset(p, 0, sizeof(*p));
    p->seed = 123;
    p->lstm_type = LSTM_TYPE_CLAMPED;
    p->n_cells = n_syms - 1;
    p->n_cells2 = p->n_cells;
    p->n_layers = 2;
    p->n_states = 10;
    p->n_streams = 2;
    p->use_layer_norm = TRUE;
    p->sgd_opt.algo = SGD_OPT_TEST;
    p->n_symbols = n_syms;

    s = lstm_init(p);

    input_syms = nc_malloc(sizeof(input_syms[0]) * s->n_states * s->n_streams);
    output_syms = nc_malloc(sizeof(output_syms[0]) * s->n_states * s->n_streams);

    coef1 = nc_find_var(s->model, coef_name);
    if (!coef1) {
        printf("Cannot find variable %s\n"
               "Available variables are:", coef_name);
        nc_dump_vars(s->model);
        exit(1);
    }
    printf("%s:\n", coef_name);

    printf("%3s %3s %10s %10s %10s %12s %12s %12s\n",
           "y", "x", "coef", "E", "dE", "grad_ref", "grad", "ratio");
    
    rnd_init(&rnd_state, 1234);

    for(it = 0; it < 10; it++) {
        coef = nc_signal_get_data(&coef_buf, coef1);
        for(i = 0; i < s->n_states * s->n_streams; i++)
            input_syms[i] = rnd_unif_u32(&rnd_state) % n_syms;
        for(i = 0; i < s->n_states * s->n_streams; i++)
            output_syms[i] = rnd_unif_u32(&rnd_state) % n_syms;

        coefs_y = rnd_unif_u32(&rnd_state) % coef->dims[0];
        if (coef->n_dims >= 2)
            coefs_x = rnd_unif_u32(&rnd_state) % coef->dims[1];
        else
            coefs_x = 0;
        pcoef = (float *)coef->data + coefs_y + coefs_x * coef->stride;
        pgrad = nc_signal_get_grad(coef1) + coefs_y + coefs_x * coef->stride;

        /* (E(c+dc)-E(c-dc)) / (2 * dc) */
        /* compute the error */
        saved_coef = *pcoef;
        *pcoef = saved_coef + eps;
        E0 = lstm_eval_error(s, input_syms, output_syms);
        *pcoef = saved_coef - eps;
        E1 = lstm_eval_error(s, input_syms, output_syms);
        /* estimate the gradient */
        grad_est = (E0 - E1) / (2 * eps);

        /* restore the coefficient */
        *pcoef = saved_coef;

        E0 = lstm_eval_error(s, input_syms, output_syms);
        /* compute the gradient */
        lstm_eval_gradient(s, output_syms);
        lstm_update(s);
        grad = *pgrad;

        if (grad_est == 0.0f) {
            if (grad == 0)
                ratio = 1.0;
            else
                ratio = 0.0;
        } else {
            ratio = grad / grad_est;
        }
        if (ratio < 0.9 || ratio > 1.1 || 1) {
            printf("%3u %3u %10f %+10.2e %+10.2e %+12.4e %+12.4e %12.3f\n",
                   coefs_y,
                   coefs_x,
                   *pcoef,
                   E0,
                   E1 - E0,
                   grad_est,
                   grad,
                   ratio);
        }
    }
    nc_free(output_syms);
    nc_free(input_syms);
    lstm_end(s);
}

static void lstm_set_default_params(LSTMParams *p)
{
    memset(p, 0, sizeof(*p));
    p->seed = 123;
    p->n_streams = 16;
    p->use_layer_norm = TRUE;
    p->lstm_type = LSTM_TYPE_CLAMPED;
    p->n_cells = 352;
    p->n_layers = 4;
    p->n_embed_out = p->n_layers;
    p->n_states = 20;
    p->n_cells2 = p->n_cells;
    p->n_symbols = 256;
    
#if 0
    p->sgd_opt.algo = SGD_OPT_BASIC;
    p->sgd_opt.learning_rate = 0.02;
#endif
#if 1
    p->sgd_opt.algo = SGD_OPT_ADAM;
    p->sgd_opt.learning_rate = 4e-3;
    p->sgd_opt.u.adam.beta1 = 0;
    p->sgd_opt.u.adam.beta2 = 0.9999;
    p->sgd_opt.u.adam.eps = 1e-5;
#endif
    p->block_len = 1600000;
}

static void lstm_dump_params(FILE *f, LSTM *s, const LSTMParams *p)
{
    char buf1[32], buf2[32], buf3[32];
    uint64_t n_params, n_params_nie;
    int mat_count;
    
    fprintf(f, "cell=");
    switch(p->lstm_type) {
    case LSTM_TYPE_CLAMPED:
        fprintf(f, "LSTM-C");
        break;
    case LSTM_TYPE_TIED:
        fprintf(f, "LSTM-T");
        break;
    case LSTM_TYPE_NORMAL:
        fprintf(f, "LSTM");
        break;
    case LSTM_TYPE_GRU:
        fprintf(f, "GRU");
        break;
    default:
        abort();
    }
    fprintf(f, " n_layer=%d hidden_size=%d batch_size=%d time_steps=%d n_symb=%d ln=%d fc=%d",
            p->n_layers,
            p->n_cells, p->n_streams, p->n_states, p->n_symbols,
            p->use_layer_norm, p->full_connect);
    if (p->n_cells2 != p->n_cells)
        fprintf(f, " proj=%d", p->n_cells2);
    if (p->n_embed_out != p->n_layers)
        fprintf(f, " n_embed_out=%d", p->n_embed_out);
    switch(p->sgd_opt.algo) {
    case SGD_OPT_BASIC:
        fprintf(f, " sgd_opt=%s lr=%0.3e", 
               "none", p->sgd_opt.learning_rate);
        break;
    case SGD_OPT_ADAM:
        fprintf(f, " sgd_opt=%s lr=%0.3e beta1=%f beta2=%f eps=%0.3e",
               "adam",
               p->sgd_opt.learning_rate,
               p->sgd_opt.u.adam.beta1,
               p->sgd_opt.u.adam.beta2,
               p->sgd_opt.u.adam.eps);
        break;
    default:
        abort();
    }

    n_params = nc_get_param_count(s->model);
    /* without input embeddings */
    n_params_nie = n_params;
    if (p->lstm_type == LSTM_TYPE_TIED || p->lstm_type == LSTM_TYPE_GRU)
        mat_count = 3;
    else
        mat_count = 4;
    n_params_nie -= p->n_layers * mat_count * p->n_cells2 * p->n_symbols;

    fprintf(f, " n_params=%s n_params_nie=%s mem=%sB\n",
            get_si_prefix(buf1, sizeof(buf1), n_params),
            get_si_prefix(buf3, sizeof(buf3), n_params_nie),
            get_si_prefix(buf2, sizeof(buf2), nc_get_heap_size(s->model)));
}

#define LSTM_FILE_MAGIC   0xb727ac57
#define LSTM_FILE_VERSION 1

static void lstm_write_file_header(FILE *f, const LSTMParams *p)
{
    fput_be32(f, LSTM_FILE_MAGIC);
    fput_be16(f, LSTM_FILE_VERSION);
    fput_u8(f, p->lstm_type);
    fput_u8(f, p->n_layers);
    fput_u8(f, p->n_embed_out);
    fput_be16(f, p->n_cells);
    fput_be16(f, p->n_states);
    fput_be16(f, p->n_streams);
    fput_be16(f, p->n_symbols);
    fput_u8(f, p->use_layer_norm | (p->full_connect << 1));
    fput_be32(f, p->seed);
    fput_sgd_opt(f, &p->sgd_opt);
}

static int lstm_read_file_header(FILE *f, LSTMParams *p)
{
    uint32_t v32;
    uint16_t v16;
    uint8_t v8;

    memset(p, 0, sizeof(*p));
    if (fget_be32(f, &v32))
        return -1;
    if (v32 != LSTM_FILE_MAGIC)
        return -1;
    if (fget_be16(f, &v16))
        return -1;
    if (v16 != LSTM_FILE_VERSION)
        return -1;
    if (fget_u8(f, &v8))
        return -1;
    p->lstm_type = v8;
    if (fget_u8(f, &v8))
        return -1;
    p->n_layers = v8;
    if (fget_u8(f, &v8))
        return -1;
    p->n_embed_out = v8;
    if (fget_be16(f, &v16))
        return -1;
    p->n_cells = v16;
    if (fget_be16(f, &v16))
        return -1;
    p->n_states = v16;
    if (fget_be16(f, &v16))
        return -1;
    p->n_streams = v16;
    if (fget_be16(f, &v16))
        return -1;
    p->n_symbols = v16;
    if (fget_u8(f, &v8))
        return -1;
    p->use_layer_norm = v8 & 1;
    p->full_connect = (v8 >> 1) & 1;
    if (fget_be32(f, &p->seed))
        return -1;
    if (fget_sgd_opt(f, &p->sgd_opt))
        return -1;
    p->n_cells2 = p->n_cells;
    return 0;
}

/* XXX: we assume a single block is written */
static void my_write_func(void *opaque, const uint8_t *buf, size_t buf_len)
{
    FILE *fo = opaque;
    fput_be32(fo, buf_len);
    fwrite(buf, 1, buf_len, fo);
}

#define STATS_DUMP_INTERVAL 100000 /* bytes */
#define STATS_AVG_COUNT (5000000 / STATS_DUMP_INTERVAL)

typedef struct {
    PutBitState pb;
} EncodeStream;

typedef uint16_t DataSymbol;

/* compress block_len bytes divided in n_streams */
void compress_block(LSTM *pred, FILE *fo,
                    const DataSymbol *block_buf, int block_len,
                    Stats *ss)
{
    EncodeStream *tab_st;
    int block_idx, c, c_in, stream_idx, cur_state, n_states, n_streams1, len;
    int n_states1, pos, block_stride, n_streams;
    int *input_syms, *output_syms;

    n_streams = pred->n_streams;
    n_states = pred->n_states;
    block_stride = (block_len + n_streams - 1) / n_streams;

    input_syms = malloc(sizeof(input_syms[0]) * n_states * n_streams);
    output_syms = malloc(sizeof(output_syms[0]) * n_states * n_streams);
    tab_st = malloc(sizeof(tab_st[0]) * n_streams);
    
    for(stream_idx = 0; stream_idx < n_streams; stream_idx++) {
        EncodeStream *st = &tab_st[stream_idx];
        put_bit_init(&st->pb, my_write_func, fo);
    }
    lstm_reset(pred);
    
    n_streams1 = min_int(block_len, n_streams);
    for(stream_idx = 0; stream_idx < n_streams1; stream_idx++) {
        EncodeStream *st = &tab_st[stream_idx];

        /* first char of each block */
        c = block_buf[stream_idx * block_stride];
        write_sym(&st->pb, pred->prob_output + stream_idx * pred->prob_output_stride, pred->n_outputs, c);
    }
    ss->n_input_bytes += n_streams1;
    block_idx = 1;
    
    /* normal batches */
    while (block_len >= (n_states * n_streams + block_idx * n_streams)) {
        prof_start(PROF_TOTAL);
        for(cur_state = 0; cur_state < n_states; cur_state++) {
            for(stream_idx = 0; stream_idx < n_streams; stream_idx++) {
                input_syms[cur_state * n_streams + stream_idx] =
                    block_buf[stream_idx * block_stride +
                              block_idx + cur_state - 1];
            }
        }
        lstm_eval(pred, -1, input_syms);
        for(cur_state = 0; cur_state < n_states; cur_state++) {
            for(stream_idx = 0; stream_idx < n_streams; stream_idx++) {
                EncodeStream *st = &tab_st[stream_idx];
                
                c = block_buf[stream_idx * block_stride +
                              block_idx + cur_state];
                prof_start(PROF_WRITE_SYM);
                write_sym(&st->pb, pred->prob_output +
                          (cur_state * n_streams + stream_idx) *
                          pred->prob_output_stride, pred->n_outputs, c);
                prof_end(PROF_WRITE_SYM);
                output_syms[cur_state * n_streams + stream_idx] = c;
            }
        }
        lstm_eval_gradient(pred, output_syms);
        lstm_update(pred);
        block_idx += n_states;
        prof_end(PROF_TOTAL);
        
        /* statistics */
        ss->n_input_bytes += n_states * n_streams;
        if ((ss->n_input_bytes - ss->last_n_input_bytes) >=
            STATS_DUMP_INTERVAL) {
            double n_output_bits = ftell(fo) * 8.0;
            for(stream_idx = 0; stream_idx < n_streams; stream_idx++) {
                EncodeStream *st = &tab_st[stream_idx];
                n_output_bits += put_bit_get_bit_count(&st->pb);
            }
            stats_dump(ss, pred->model, n_output_bits);
        }
    }
    
    len = block_len - block_idx * n_streams;
    if (len > 0) {
        ss->n_input_bytes += len;
        /* no training for the last batch */
        n_states1 = (len + n_streams - 1) / n_streams;

        for(cur_state = 0; cur_state < n_states; cur_state++) {
            for(stream_idx = 0; stream_idx < n_streams; stream_idx++) {
                if (cur_state < n_states1) {
                    pos = stream_idx * block_stride + block_idx + cur_state - 1;
                    if (pos < block_len)
                        c_in = block_buf[pos];
                    else
                        c_in = 0;
                } else {
                    c_in = 0;
                }
                input_syms[cur_state * n_streams + stream_idx] = c_in;
            }
        }
        lstm_eval(pred, -1, input_syms);
        for(cur_state = 0; cur_state < n_states1; cur_state++) {
            for(stream_idx = 0; stream_idx < n_streams; stream_idx++) {
                EncodeStream *st = &tab_st[stream_idx];
                pos = stream_idx * block_stride + block_idx + cur_state;
                if (pos < block_len) {
                    c = block_buf[pos];
                    write_sym(&st->pb, pred->prob_output +
                              (cur_state * n_streams + stream_idx) *
                              pred->prob_output_stride, pred->n_outputs, c);
                }
            }
        }
    }
    
    /* file output */
    fput_be32(fo, block_len);
    for(stream_idx = 0; stream_idx < n_streams; stream_idx++) {
        EncodeStream *st = &tab_st[stream_idx];
        put_bit_end(&st->pb);
    }
    fflush(fo);
    free(tab_st);
    free(input_syms);
    free(output_syms);
}

static const char *load_coefs_filename;
static BOOL quiet_flag;

void compress_file(const LSTMParams *p,
                   const char *in_filename, const char *out_filename,
                   const char *debug_dir)
{
    FILE *f, *fo;
    LSTM *pred;
    Stats st_s, *st = &st_s;
    DataSymbol *block_buf;
    int block_len, n_streams, n_states, block_len_max, symb_shift, c;
    int64_t n_output_bytes;
    char out_filename_buf[1024];
    
    pred = lstm_init(p);
    if (!quiet_flag) {
        lstm_dump_params(stdout, pred, p);
    }

    if (load_coefs_filename)
        nc_load_coefs(pred->model, load_coefs_filename);
    
    f = fopen(in_filename, "rb");
    if (!f) {
        perror(in_filename);
        exit(1);
    }

    n_streams = pred->n_streams;
    n_states = pred->n_states;
    
    block_len_max = (((p->block_len - 1) / (n_states * n_streams)) *
                     n_states + 1) * n_streams;
    block_buf = nc_malloc(block_len_max * sizeof(block_buf[0]));
    
    stats_init(st, debug_dir, "nncp", in_filename, STATS_AVG_COUNT);
    if (st->fp) {
        fprintf(st->fp, "# ");
        lstm_dump_params(st->fp, pred, p);
    }
    
    if (!out_filename) {
        snprintf(out_filename_buf, sizeof(out_filename_buf), "%s/%s",
                 st->debug_dir, "out.bin");
        out_filename = out_filename_buf;
    }
    st->learning_rate = p->sgd_opt.learning_rate;
        
    fo = fopen(out_filename, "wb");
    if (!fo) {
        perror(out_filename);
        exit(1);
    }

    lstm_write_file_header(fo, p);

    if (p->n_symbols <= 256)
        symb_shift = 0;
    else
        symb_shift = 1;
        
    for(;;) {
        for(block_len = 0; block_len < block_len_max; block_len++) {
            if (symb_shift == 0) {
                c = fgetc(f);
                if (c < 0)
                    break;
            } else {
                uint16_t v16;
                if (fget_be16(f, &v16))
                    break;
                c = v16;
            }
            if (c >= p->n_symbols) {
                fprintf(stderr, "Invalid symbol: %d\n", c);
                exit(1);
            }
            block_buf[block_len] = c;
        }
        if (block_len == 0)
            break;
        compress_block(pred, fo, block_buf, block_len, st);
    }
    nc_free(block_buf);
    
    n_output_bytes = ftell(fo);
    fclose(fo);
    fclose(f);

    stats_end(st, pred->model, n_output_bytes);
    
    lstm_end(pred);

    nc_prof_dump();
}

typedef struct {
    uint8_t *buf;
    GetBitState gb;
} DecodeStream;

void decompress_file(const char *in_filename, const char *out_filename)
{
    LSTMParams p_s, *p = &p_s;
    FILE *f, *fo;
    LSTM *pred;
    int64_t byte_pos, ti;
    uint32_t block_len, cblock_len;
    int c, n_streams, n_states, block_stride, stream_idx, n_states1, n_streams1;
    int len, byte_pos1, block_idx, cur_state, c_in, symb_shift, i;
    int *input_syms, *output_syms, pos;
    DataSymbol *block_buf;
    DecodeStream *dec_st;
    
    f = fopen(in_filename, "rb");
    if (!f) {
        perror(in_filename);
        exit(1);
    }

    if (lstm_read_file_header(f, p) < 0) {
        fprintf(stderr, "Invalid file header\n");
        exit(1);
    }
    
    fo = fopen(out_filename, "wb");
    if (!fo) {
        perror(out_filename);
        exit(1);
    }

    p->is_decoder = TRUE;
    pred = lstm_init(p);
    if (!quiet_flag) {
        lstm_dump_params(stdout, pred, p);
    }

    if (load_coefs_filename)
        nc_load_coefs(pred->model, load_coefs_filename);

    if (p->n_symbols <= 256)
        symb_shift = 0;
    else
        symb_shift = 1;

    n_streams = pred->n_streams;
    n_states = pred->n_states;
    input_syms = malloc(sizeof(input_syms[0]) * n_streams);
    output_syms = malloc(sizeof(output_syms[0]) * n_states * n_streams);
    dec_st = nc_mallocz(sizeof(dec_st[0]) * n_streams);
    byte_pos = 0;
    byte_pos1 = 0;

    ti = get_time_ms();
    for(;;) {
        if (fget_be32(f, &block_len) < 0)
            break;
        if (block_len == 0)
            break;
        block_stride = (block_len + n_streams - 1) / n_streams;
        
        block_buf = nc_malloc(block_stride * n_streams * sizeof(block_buf[0]));
        for(stream_idx = 0; stream_idx < n_streams; stream_idx++) {
            DecodeStream *st = &dec_st[stream_idx];
            if (fget_be32(f, &cblock_len) < 0) {
                fprintf(stderr, "Could not read cblock_len\n");
                exit(1);
            }
            st->buf = nc_malloc(cblock_len);
            if (fread(st->buf, 1, cblock_len, f) != cblock_len) {
                fprintf(stderr, "Could not read input stream\n");
                exit(1);
            }
            get_bit_init(&st->gb, st->buf, cblock_len, NULL, NULL);
        }
        lstm_reset(pred);


        /* first byte */
        n_streams1 = min_int(block_len, n_streams);
        for(stream_idx = 0; stream_idx < n_streams1; stream_idx++) {
            DecodeStream *st = &dec_st[stream_idx];
            c = read_sym(&st->gb, pred->prob_output + stream_idx * pred->prob_output_stride, pred->n_outputs);
            block_buf[stream_idx * block_stride] = c;
        }
        
        block_idx = 1;
        /* normal batches */
        while (block_len >= (n_states * n_streams + block_idx * n_streams)) {
            prof_start(PROF_TOTAL);
            for(cur_state = 0; cur_state < n_states; cur_state++) {
                for(stream_idx = 0; stream_idx < n_streams; stream_idx++) {
                    input_syms[stream_idx] = 
                        block_buf[stream_idx * block_stride +
                                  block_idx + cur_state - 1];
                }
                lstm_eval(pred, cur_state, input_syms);
                for(stream_idx = 0; stream_idx < n_streams; stream_idx++) {
                    DecodeStream *st = &dec_st[stream_idx];
                    c = read_sym(&st->gb, pred->prob_output +
                                 (cur_state * n_streams + stream_idx) *
                                 pred->prob_output_stride, pred->n_outputs);
                    //                    printf("%d.%d: c=%d\n", cur_state, stream_idx, c);
                    block_buf[stream_idx * block_stride +
                              block_idx + cur_state] = c;
                    output_syms[cur_state * n_streams + stream_idx] = c;
                }
            }
            lstm_eval_gradient(pred, output_syms);
            lstm_update(pred);
            block_idx += n_states;
            prof_end(PROF_TOTAL);
        }

        /* last partial batch */
        len = block_len - block_idx * n_streams;
        if (len > 0) {
            /* no training for the last batch */
            n_states1 = (len + n_streams - 1) / n_streams;
            for(cur_state = 0; cur_state < n_states1; cur_state++) {
                for(stream_idx = 0; stream_idx < n_streams; stream_idx++) {
                    pos = stream_idx * block_stride + block_idx + cur_state - 1;
                    if (pos < block_len)
                        c_in = block_buf[pos];
                    else
                        c_in = 0;
                    input_syms[stream_idx] = c_in;
                }
                lstm_eval(pred, cur_state, input_syms);

                for(stream_idx = 0; stream_idx < n_streams; stream_idx++) {
                    DecodeStream *st = &dec_st[stream_idx];
                    pos = stream_idx * block_stride + block_idx + cur_state;
                    if (pos < block_len) {
                        c = read_sym(&st->gb, pred->prob_output +
                                     (cur_state * n_streams + stream_idx) *
                                     pred->prob_output_stride,
                                     pred->n_outputs);
                        block_buf[pos] = c;
                    }
                }
            }
        }

        for(i = 0; i < block_len; i++) {
            if (symb_shift == 0)
                fputc(block_buf[i], fo);
            else
                fput_be16(fo, block_buf[i]);
        }
        
        nc_free(block_buf);

        for(stream_idx = 0; stream_idx < n_streams; stream_idx++) {
            DecodeStream *st = &dec_st[stream_idx];
            nc_free(st->buf);
        }

        byte_pos += block_len;
        byte_pos1 += block_len;
        if (byte_pos1 >= 100000) {
            if (!quiet_flag) {
                printf("%12" PRIu64 "\r", byte_pos);
                fflush(stdout);
            }
            byte_pos1 = 0;
        }
    }
    ti = get_time_ms() - ti;

    if (!quiet_flag) {
        printf("%12" PRIu64 " (%0.2f kS/s)\n", byte_pos, (double)byte_pos / ti);
    }
    
    fclose(fo);
    fclose(f);
    nc_free(dec_st);
    free(input_syms);
    free(output_syms);
    lstm_end(pred);

    nc_prof_dump();
}

static struct option options[] = {
    { "full_connect", required_argument },
    { "n_symb", required_argument },
    { "load_coefs", required_argument },

    { "cell", required_argument },
    { "hidden_size", required_argument },
    { "n_layer", required_argument },
    { "time_steps", required_argument },
    { "batch_size", required_argument },
    { "layer_norm", required_argument },
    { "seed", required_argument },
    { "sgd_opt", required_argument },
    { "lr", required_argument },
    { "adam_beta1", required_argument },
    { "adam_beta2", required_argument },
    { "adam_eps", required_argument },
    { "block_len", required_argument },
    { "n_embed_out", required_argument },
    { NULL },
};

static void help(void)
{
    printf("NNCP version " CONFIG_VERSION", Copyright (c) 2018-2019 Fabrice Bellard\n"
           "Lossless data compression with LSTM\n"
           "usage: nncp [options] c infile outfile\n"
           "                      d infile outfile\n"
           "\n"
           "General options:\n"
           "-h                            show this help\n"
           "-T n                          number of threads (default=1)\n"
           "-d path                       save logs & coefficients\n"
           "-q                            less verbose output\n"
           "\n"
           "RNN Model options (only for compression):\n"
           "-cell [lstm|lstmc|lstmt]      LSTM cell variant\n"
           "-hidden_size n                number of LSTM hidden states\n"
           "-n_layer n                    number of layers\n"
           "-time_steps n                 number of time steps for TBTT\n"
           "-batch_size n                 batch size\n"
           "-layer_norm [0|1]             enable layer normalization\n"
           "-seed n                       random number generator seed\n"
           "-sgd_opt [none|adam]          SGD optimization algorithm\n"
           "-lr n                         learning rate\n"
           "-adam_beta1 n                 ADAM beta1 parameter\n"
           "-adam_beta2 n                 ADAM beta2 parameter\n"
           "-adam_eps                     ADAM epsilon parameter\n"
           "-block_len n                  encoded block length\n"
           "-full_connect [0|1]           fully connect all the layers\n"
           "-n_embed_out n                number of layers in output embedding\n"
           "\n"
           "Debug options:\n"
           "-load_coefs file     load coefficients from file\n"
           );
    exit(1);
}

int main(int argc, char **argv)
{
    int c, option_index;
    const char *debug_dir = NULL;
    LSTMParams p_s, *p = &p_s;

    lstm_set_default_params(p);
    debug_dir = NULL;
    for(;;) {
        c = getopt_long_only(argc, argv, "hqT:d:", options, &option_index);
        if (c == -1)
            break;
        switch(c) {
        case 0:
            switch(option_index) {
            case 0:
                p->full_connect = (atoi(optarg) != 0);
                break;
            case 1:
                p->n_symbols = atoi(optarg);
                if (p->n_symbols < 2 || p->n_symbols > 65536) {
                    fprintf(stderr, "invalid number of symbols\n");
                    exit(1);
                }
                break;
            case 2: /* load_coefs */
                load_coefs_filename = optarg;
                break;
            case 3:
                if (!strcmp(optarg, "lstm"))
                    p->lstm_type = LSTM_TYPE_NORMAL;
                else if (!strcmp(optarg, "lstmc"))
                    p->lstm_type = LSTM_TYPE_CLAMPED;
                else if (!strcmp(optarg, "lstmt"))
                    p->lstm_type = LSTM_TYPE_TIED;
#if 0
                /* GRU is no longer supported */
                else if (!strcmp(optarg, "gru"))
                    p->lstm_type = LSTM_TYPE_GRU;
#endif
                else {
                    fprintf(stderr, "Unknown cell type: %s\n", optarg);
                    exit(1);
                }
                break;
            case 4:
                p->n_cells2 = p->n_cells = atoi(optarg);
                break;
            case 5:
                p->n_layers = atoi(optarg);
                p->n_embed_out = p->n_layers;
                break;
            case 6:
                p->n_states = atoi(optarg);
                break;
            case 7:
                p->n_streams = atoi(optarg);
                break;
            case 8:
                p->use_layer_norm = (atoi(optarg) != 0);
                break;
            case 9:
                p->seed = strtoul(optarg, NULL, 0);
                break;
            case 10:
                if (!strcmp(optarg, "none")) {
                    p->sgd_opt.algo = SGD_OPT_BASIC;
                } else if (!strcmp(optarg, "adam")) {
                    p->sgd_opt.algo = SGD_OPT_ADAM;
                } else {
                    fprintf(stderr, "Unknown SGD optimization: %s\n", optarg);
                    exit(1);
                }
                break;
            case 11:
                p->sgd_opt.learning_rate = strtod(optarg, NULL);
                break;
            case 12:
                p->sgd_opt.u.adam.beta1 = strtod(optarg, NULL);
                break;
            case 13:
                p->sgd_opt.u.adam.beta2 = strtod(optarg, NULL);
                break;
            case 14:
                p->sgd_opt.u.adam.eps = strtod(optarg, NULL);
                break;
            case 15:
                p->block_len = atoi(optarg);
                break;
            case 16:
                p->n_embed_out = atoi(optarg);
                break;
            default:
                fprintf(stderr, "unknown option index: %d\n", option_index);
                exit(1);
            }
            break;
        case 'h':
            help();
            break;
        case 'q':
            quiet_flag = TRUE;
            break;
        case 'T':
            nb_threads = atoi(optarg);
            break;
        case 'd':
            debug_dir = optarg;
            break;
        default:
            exit(1);
        }
    }
            
    if (optind >= argc) 
        help();

    if (argv[optind][0] == 'c') {
        const char *out_filename;
        if (optind + 1 >= argc) 
            help();
        if ((optind + 2) < argc) {
            out_filename = argv[optind + 2];
        } else {
            out_filename = NULL;
            if (!debug_dir)
                help();
        }
        compress_file(p, argv[optind + 1], out_filename, debug_dir);
    } else if (argv[optind][0] == 'd') {
        if (optind + 2 >= argc) 
            help();
        decompress_file(argv[optind + 1], argv[optind + 2]);
    } else if (argv[optind][0] == 't') {
        float eps;
        const char *var_name;
        if (optind + 1 >= argc) 
            help();
        var_name = argv[optind + 1];
        eps = 0.1;
        if ((optind + 2) < argc)
            eps = strtof(argv[optind + 2], NULL);
        gradient_test(var_name, eps);
    } else {
        help();
    }
    return 0;
}
