/*
 * Lossless compression with Transformer
 * 
 * Copyright (c) 2018-2019 Fabrice Bellard
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <inttypes.h>
#include <assert.h>
#include <time.h>
#include <getopt.h>
#include <stdarg.h>
#include <sys/time.h>

#include "cutils.h"
#include "arith.h"
#include "libnc.h"
#include "cp_utils.h"

/************************************************/
/* Transformer model */

//#define USE_CONCAT_EMBED

#define LR_MAX_STEPS 3

typedef struct {
    BOOL is_decoder;
    int n_layer;
    int d_model;
    int n_head;
    int d_key;
    int d_value;
    int d_inner;
    int d_pos;
    int mem_len;
    int train_len;
    int n_symbols;
    uint8_t tied_embed;
    uint32_t seed;
    float init_range;
    float init_bias_v;
    SGDOptParams sgd_opt;
    BOOL more_relu;
    /* learning rate schedule */
    int lr_steps;
    float lr[LR_MAX_STEPS + 1];
    int64_t lr_pos[LR_MAX_STEPS];
} TransformerModelParams;

typedef struct {
    RNDState rnd_state;
    NCModel *model;
    int n_layer;
    int d_model;
    int n_head;
    int d_key;
    int d_value;
    int d_inner;
    int d_pos;
    int mem_len;
    int train_len;
    int n_symbols;
    NCSignal **mem_h; /* n_layer */
    NCSignal **train_h; /* n_layer */
    NCSignal *input; /* vector of length train_len */
    NCSignal *output; /* matrix of n_symbols * train_len */
    NCSignal *expected_output;

    int n_streams; /* rename */
    int n_states; /* same as train_len -> remove */
    size_t prob_output_stride;
    float *prob_output;
    NCCompileResult func;
    NCSGDOptState *sgd_opt;
} TransformerModel;

typedef struct {
    NCSignal *w_q, *w_k, *w_v; /* self attention matrixes */
    NCSignal *w_o; /* linear output matrix */
    NCSignal *g1, *b1; /* layer norm after w_o */
    NCSignal *ff1, *ff2, *ff_bias1; /* feed forward layer */
    NCSignal *g2, *b2; /* layer norm after ff2 */
} TransformerLayer;

static NCSignal *var_init(NCSignal *x, float range)
{
    return nc_set_var_init(x, 0, range);
}

static int nb_threads = 1;

static TransformerModel *trf_init(const TransformerModelParams *p)
{
    TransformerModel *s;
    NCModel *m;
    int layer_idx, n;
    TransformerLayer *layers, *tl;
    NCSignal *embed, *embed_out, *layer_input, **tab_tmp, *out_bias;
    NCSignal *w_r; /* learned relative position for each head */
    NCSignal *bias_q; /* query bias for relative position */
    NCSignal *bias_v; /* query bias for relative position */
    
    s = nc_mallocz(sizeof(*s));
    rnd_init(&s->rnd_state, p->seed);
    s->n_layer = p->n_layer;
    s->d_model = p->d_model;
    s->n_head = p->n_head;
    s->d_key = p->d_key;
    s->d_value = p->d_value;
    s->d_inner = p->d_inner;
    s->d_pos = p->d_pos;
    s->mem_len = p->mem_len;
    s->train_len = p->train_len;
    s->n_symbols = p->n_symbols;

    s->n_streams = 1;
    s->n_states = s->train_len;
    
    m = nc_model_init(nb_threads);
    s->model = m;
    nc_start_func(m, "eval");

    w_r = nc_signal_set_name(nc_new_var_3d(m, NC_TYPE_F32,
                                           s->d_pos, s->d_key, s->n_head),
                             "w_r");
    var_init(w_r, p->init_range);
                
    bias_q = 
        nc_signal_set_name(nc_new_var_1d(m, NC_TYPE_F32, s->n_head * s->d_value),
                           "bias_q");
    bias_v = 
        nc_signal_set_name(nc_new_var_1d(m, NC_TYPE_F32, s->n_head * s->d_value),
                           "bias_v");
    var_init(bias_v, p->init_bias_v);
    
    layers = nc_mallocz(sizeof(layers[0]) * s->n_layer);
    for(layer_idx = 0; layer_idx < s->n_layer; layer_idx++) {
        tl = &layers[layer_idx];
        tl->w_q = 
            nc_signal_set_name(nc_new_var_2d(m, NC_TYPE_F32, s->n_head * s->d_key,
                                                     s->d_model),
                               "w_q_%d", layer_idx);
        var_init(tl->w_q, p->init_range);
        tl->w_k =
            nc_signal_set_name(nc_new_var_2d(m, NC_TYPE_F32, s->n_head * s->d_key,
                                                     s->d_model),
                               "w_k_%d", layer_idx);
        var_init(tl->w_k, p->init_range);
        tl->w_v =
            nc_signal_set_name(nc_new_var_2d(m, NC_TYPE_F32, s->n_head * s->d_value,
                                                     s->d_model),
                               "w_v_%d", layer_idx);
        var_init(tl->w_v, p->init_range);
        tl->w_o = 
            nc_signal_set_name(nc_new_var_2d(m, NC_TYPE_F32, s->d_model,
                                                     s->n_head * s->d_value),
                               "w_o_%d", layer_idx);
        var_init(tl->w_o, p->init_range);
        tl->g1 = 
            nc_signal_set_name(nc_new_var_1d(m, NC_TYPE_F32, s->d_model),
                               "g1_%d", layer_idx);
        nc_set_var_init(tl->g1, 1, 0);
        tl->b1 = 
            nc_signal_set_name(nc_new_var_1d(m, NC_TYPE_F32, s->d_model),
                               "b1_%d", layer_idx);
        tl->ff1 = 
            nc_signal_set_name(nc_new_var_2d(m, NC_TYPE_F32, s->d_inner,
                                              s->d_model),
                               "ff1_%d", layer_idx);
        var_init(tl->ff1, p->init_range);

        tl->ff_bias1 = 
            nc_signal_set_name(nc_new_var_1d(m, NC_TYPE_F32, s->d_inner),
                               "ff_bias1_%d", layer_idx);

        tl->ff2 = 
            nc_signal_set_name(nc_new_var_2d(m, NC_TYPE_F32, s->d_model,
                                                     s->d_inner),
                               "ff2_%d", layer_idx);
        var_init(tl->ff2, p->init_range);

        tl->g2 = 
            nc_signal_set_name(nc_new_var_1d(m, NC_TYPE_F32, s->d_model),
                               "g2_%d", layer_idx);
        nc_set_var_init(tl->g2, 1, 0);
        tl->b2 = 
            nc_signal_set_name(nc_new_var_1d(m, NC_TYPE_F32, s->d_model),
                               "b2_%d", layer_idx);
    }
    embed = nc_signal_set_name(nc_new_var_2d(m, NC_TYPE_F32, s->d_model,
                                              s->n_symbols),
                               "embed");
    var_init(embed, p->init_range);
#ifdef USE_CONCAT_EMBED
    n = s->d_model * s->n_layer;
#else
    n = s->d_model;
#endif
    if (p->tied_embed) {
        embed_out = NULL;
    } else {
        embed_out = 
            nc_signal_set_name(nc_new_var_2d(m, NC_TYPE_F32, s->n_symbols, n),
                               "embed_out");
        var_init(embed_out, p->init_range);
    }
    
    out_bias = nc_signal_set_name(nc_new_var_1d(m, NC_TYPE_F32, s->n_symbols),
                                  "out_bias");
        
    tab_tmp = nc_mallocz(sizeof(tab_tmp[0]) *
                          max_int(max_int(2, s->train_len),
                                  max_int(s->n_head, s->n_layer)));

    s->mem_h = nc_mallocz(sizeof(s->mem_h[0]) * s->n_layer);
    for(layer_idx = 0; layer_idx < s->n_layer; layer_idx++) {
        NCSignal *x;
        x = nc_new_local_2d(m, NC_TYPE_F32, s->d_model, s->mem_len);
        nc_signal_set_name(x, "mem_h_%d", layer_idx);
        nc_signal_set_global(x);
        s->mem_h[layer_idx] = x;
    }

    s->train_h = nc_mallocz(sizeof(s->train_h[0]) * s->n_layer);

    s->input = nc_new_local_1d(m, NC_TYPE_I32, s->train_len);
    nc_signal_set_name(s->input, "input");
    nc_signal_set_global(s->input);
    
    layer_input = nc_get_col(embed, s->input);

    for(layer_idx = 0; layer_idx < s->n_layer; layer_idx++) {
        NCSignal *t0, *query, *key, *value;
        NCSignal *ff_input, *rd, *query1, *query2;
        
        tl = &layers[layer_idx];

        nc_signal_set_name(layer_input, "train_h_%d", layer_idx);
        nc_signal_set_global(layer_input);
        s->train_h[layer_idx] = layer_input;
        
        tab_tmp[0] = s->mem_h[layer_idx];
        tab_tmp[1] = layer_input;
        t0 = nc_hconcat(tab_tmp, 2);

        query = nc_matmul(tl->w_q, layer_input);
        key = nc_matmul(tl->w_k, t0);
        value = nc_matmul(tl->w_v, t0);

        /* split query, key and value for each head */
        key = nc_vsplit_dim(key, s->n_head);
        /* add the bias for relative position for query */
        t0 = nc_add(query, bias_q);
        query1 = nc_vsplit_dim(t0, s->n_head);

        t0 = nc_add(query, bias_v);
        query2 = nc_vsplit_dim(t0, s->n_head);
        
        value = nc_vsplit_dim(value, s->n_head);
        
        /* cross product term */
        t0 = nc_matmul_hint(key, query1,
                            TRUE, FALSE, NC_MATMUL_HINT1);
        
        /* relative distance term */
        rd = nc_matmul(w_r, query2);
        rd = nc_pad(rd, s->mem_len + s->train_len - s->d_pos, NC_PAD_DUP,
                    0, NC_PAD_ZERO);
        rd = nc_rel_shift(rd, s->train_len - 1);
        
        t0 = nc_add(rd, t0);
        t0 = nc_mul_const(t0, 1.0f / sqrtf(s->d_key));

        /* set the future cross products to -infinity so that they
           don't change the softmax result */
        t0 = nc_slt_mat_set(t0, s->mem_len + 1, -INFINITY);
        
        t0 = nc_soft_max(t0);
        t0 = nc_matmul_hint(value, t0, FALSE, FALSE, NC_MATMUL_HINT2);

        /* merge all the heads */
        t0 = nc_vconcat_dim(t0);

        /* linear layer */
        t0 = nc_matmul(tl->w_o, t0);

        t0 = nc_add(t0, layer_input);
        /* layer norm 1 */
        t0 = nc_add(nc_mul(nc_layer_norm(t0),
                           nc_dup(tl->g1, s->train_len)),
                    tl->b1);
        if (p->more_relu) {
            t0 = nc_relu(t0);
        }
        
        ff_input = t0;
        
        t0 = nc_add(nc_matmul(tl->ff1, t0), tl->ff_bias1);
        t0 = nc_relu(t0);
        
        t0 = nc_matmul(tl->ff2, t0);
        
        t0 = nc_add(t0, ff_input);
        /* layer norm 2 */
        layer_input = nc_add(nc_mul(nc_layer_norm(t0),
                                    nc_dup(tl->g2, s->train_len)),
                             tl->b2);
    }
    
    {
        NCSignal *t0;
        s->expected_output = nc_new_local_1d(m, NC_TYPE_I32, s->train_len);
        nc_signal_set_name(s->expected_output, "eoutput");
        nc_signal_set_global(s->expected_output);

        if (p->tied_embed) {
            t0 = layer_input;
            if (p->tied_embed == 2) {
                NCSignal *w;
                /* add a projection */
                w = nc_signal_set_name(nc_new_var_2d(m, NC_TYPE_F32, s->d_model,
                                                      s->d_model),
                                       "p");
                nc_set_var_init(w, 1.0 / sqrtf(s->d_model), 0);
                t0 = nc_matmul(w, t0);
            }
            t0 = nc_matmul(nc_transpose(embed), t0);
        } else {
#ifdef USE_CONCAT_EMBED
            for(i = 0; i < s->n_layer - 1; i++)
                tab_tmp[i] = s->train_h[i + 1];
            tab_tmp[s->n_layer - 1] = layer_input;
            t0 = nc_vconcat(tab_tmp, s->n_layer);
#else
            t0 = layer_input;
#endif
            t0 = nc_matmul(embed_out, t0);
        }
        t0 = nc_add(t0, out_bias);

        s->output = nc_log_soft_max(t0, s->expected_output);
        nc_signal_set_name(s->output, "output");
        nc_signal_set_global(s->output);
    }
    
    for(layer_idx = 0; layer_idx < s->n_layer; layer_idx++) {
        tl = &layers[layer_idx];
    }
    nc_free(layers);
    nc_free(tab_tmp);

    {
        NCCompileParams sp_s, *sp = &sp_s;
        memset(sp, 0, sizeof(*sp));
        sp->grad = TRUE;
        if (p->is_decoder) {
            sp->partial_eval = TRUE;
            sp->output = s->output;
            sp->output_axis = 1;
        }
        nc_compile(m, &s->func, sp);
    }

    nc_var_init(m, &s->rnd_state);

    s->sgd_opt = nc_sgd_opt_init(m, &p->sgd_opt);

    /* apply the SGD optimizer to all the parameters */
    nc_sgd_opt_set_all(m, s->sgd_opt);

    {
        NCSignalData xbuf, *x;
        x = nc_signal_get_data(&xbuf, s->output);
        s->prob_output = x->data;
        s->prob_output_stride = x->stride;
    }
    return s;
}

static void trf_end(TransformerModel *s)
{
    nc_free(s->mem_h);
    nc_free(s->train_h);
    nc_free(s->func.partial_eval_func);
    nc_model_end(s->model);
    nc_free(s);
}

static void trf_reset(TransformerModel *s)
{
    int stream_idx;
    float a;
    int i;
    
    for(stream_idx = 0; stream_idx < s->n_streams; stream_idx++) {
        float *output = s->prob_output + stream_idx * s->prob_output_stride;
        /* for the first char */
        a = 1.0f / s->n_symbols;
        for(i = 0; i < s->n_symbols; i++)
            output[i] = a;
    }
}

static void trf_eval(TransformerModel *s,
                     int output_index,
                     const int *input_syms)
{
    int i;
    NCSignalData xbuf, *x;
    
    prof_start(PROF_EVAL);

    if (output_index < 0) {
        for(i = 0; i < s->train_len; i++) {
            x = nc_signal_get_data(&xbuf, s->input);
            ((int32_t *)x->data)[i] = input_syms[i];
        }
        nc_eval(s->func.eval_func);
    } else {
        x = nc_signal_get_data(&xbuf, s->input);
        ((int32_t *)x->data)[output_index] = input_syms[output_index];
        
        nc_eval(s->func.partial_eval_func[output_index]);
    }
    
    prof_end(PROF_EVAL);
}

/* evaluate the gradient */
static void trf_eval_gradient(TransformerModel *s, const int *output_syms)
{
    int i;

    prof_start(PROF_GRAD);
    for(i = 0; i < s->train_len; i++) {
        NCSignalData xbuf, *x = nc_signal_get_data(&xbuf, s->expected_output);
        int32_t *data = (int32_t *)x->data;
        data[i] = output_syms[i];
    }
    nc_eval(s->func.grad_func);
    prof_end(PROF_GRAD);
}

static void mem_update(TransformerModel *s, NCSignal **mem, NCSignal **train)
{
    size_t i, j;
    NCSignalData *x, *y, xbuf, ybuf;

    for(i = 0; i < s->n_layer; i++) {
        x = nc_signal_get_data(&xbuf, mem[i]);
        y = nc_signal_get_data(&ybuf, train[i]);
        assert(x->dims[0] == y->dims[0]);
        assert(x->dims[1] == s->mem_len);
        assert(y->dims[1] == s->train_len);

        if (s->mem_len > s->train_len) {
            for(j = 0; j < s->mem_len - s->train_len; j++) {
                memcpy((float *)x->data + j * x->stride,
                       (float *)x->data + (j + s->train_len) * x->stride,
                       x->dims[0] * sizeof(float));
            }
            for(j = 0; j < s->train_len; j++) {
                memcpy((float *)x->data + (s->mem_len - s->train_len + j) * x->stride,
                       (float *)y->data + j * y->stride,
                       x->dims[0] * sizeof(float));
            }
        } else {
            for(j = 0; j < s->mem_len; j++) {
                memcpy((float *)x->data + j * x->stride,
                       (float *)y->data + (s->train_len - s->mem_len + j) * y->stride,
                       x->dims[0] * sizeof(float));
            }
        }
    }
}

/* update with coefficients */
static void trf_update(TransformerModel *s)
{
    prof_start(PROF_UPDATE);
    nc_update(s->model);

    /* shift the memory */
    mem_update(s, s->mem_h, s->train_h);
    
    prof_end(PROF_UPDATE);
}

void trf_dump_params(FILE *f, TransformerModel *s,
                     const TransformerModelParams *p, BOOL dump_lr)
{
    int i;
    char buf1[32], buf2[32], buf3[32];
    uint64_t n_params, n_params_nie;
    
    fprintf(f, "n_layer=%d d_model=%d n_head=%d mem_len=%d train_len=%d d_pos=%d d_inner=%d n_symb=%d tied=%d init_range=%f",
            p->n_layer,
            p->d_model, 
            p->n_head, 
            p->mem_len, 
            p->train_len,
            p->d_pos,
            p->d_inner,
            p->n_symbols,
            p->tied_embed,
            p->init_range);
    if (dump_lr) {
        fprintf(f, " lr=%0.2e", p->lr[0]);
        for(i = 0; i < p->lr_steps; i++) {
            fprintf(f, ",%" PRIu64 ",%0.2e", p->lr_pos[i], p->lr[i + 1]);
        }
    }
    n_params = nc_get_param_count(s->model);
    /* without input embeddings */
    n_params_nie = n_params;
    if (!p->tied_embed)
        n_params_nie -= p->d_model * p->n_symbols;

    fprintf(f, " n_params=%s n_params_nie=%s mem=%sB\n",
            get_si_prefix(buf1, sizeof(buf1), n_params),
            get_si_prefix(buf3, sizeof(buf3), n_params_nie),
            get_si_prefix(buf2, sizeof(buf2), nc_get_heap_size(s->model)));
}

void trf_set_default_params(TransformerModelParams *p)
{
    memset(p, 0, sizeof(*p));
    p->seed = 123;
    p->n_layer = 4;
    p->d_model = 256;
    p->n_head = 8;
    p->d_key = p->d_model / p->n_head;
    p->d_value = p->d_key;
    p->d_inner = p->d_model * 2;
    p->d_pos = 16;
    p->mem_len = 64;
    p->train_len = 64;
    p->n_symbols = 256;
    p->tied_embed = 0;
    p->init_range = 0.05;
    p->init_bias_v = 1.3;
    p->more_relu = FALSE;
    
    p->sgd_opt.algo = SGD_OPT_ADAM;
    p->sgd_opt.u.adam.beta1 = 0;
    p->sgd_opt.u.adam.beta2 = 0.9999;
    p->sgd_opt.u.adam.eps = 1e-5;

    p->lr_steps = 1;
    p->lr[0] = 3e-4;
    p->lr[1] = 1e-4;
    p->lr_pos[0] = 5e6;
    p->sgd_opt.learning_rate = p->lr[0];
}

#define TRF_FILE_MAGIC   0xb727ac58
#define TRF_FILE_VERSION 1

static void trf_write_file_header(FILE *f, const TransformerModelParams *p)
{
    fput_be32(f, TRF_FILE_MAGIC);
    fput_be16(f, TRF_FILE_VERSION);
    fput_u8(f, p->n_layer);
    fput_u8(f, p->n_head);
    fput_be16(f, p->d_key);
    fput_be16(f, p->d_value);
    fput_be16(f, p->d_inner);
    fput_be16(f, p->d_pos);
    fput_be16(f, p->mem_len);
    fput_be16(f, p->train_len);
    fput_be16(f, p->n_symbols);
    fput_u8(f, p->tied_embed);
    fput_be32(f, p->seed);
    fput_f32(f, p->init_range);
    fput_f32(f, p->init_bias_v);
    fput_u8(f, p->more_relu);
    fput_sgd_opt(f, &p->sgd_opt);
}

static int trf_read_file_header(FILE *f, TransformerModelParams *p)
{
    uint32_t v32;
    uint16_t v16;
    uint8_t v8;

    memset(p, 0, sizeof(*p));
    if (fget_be32(f, &v32))
        return -1;
    if (v32 != TRF_FILE_MAGIC)
        return -1;
    if (fget_be16(f, &v16))
        return -1;
    if (v16 != TRF_FILE_VERSION)
        return -1;
    if (fget_u8(f, &v8))
        return -1;
    p->n_layer = v8;
    if (fget_u8(f, &v8))
        return -1;
    p->n_head = v8;
    if (fget_be16(f, &v16))
        return -1;
    p->d_key = v16;
    p->d_model = p->d_key * p->n_head;
    if (fget_be16(f, &v16))
        return -1;
    p->d_value = v16;
    if (fget_be16(f, &v16))
        return -1;
    p->d_inner = v16;
    if (fget_be16(f, &v16))
        return -1;
    p->d_pos = v16;
    if (fget_be16(f, &v16))
        return -1;
    p->mem_len = v16;
    if (fget_be16(f, &v16))
        return -1;
    p->train_len = v16;
    if (fget_be16(f, &v16))
        return -1;
    p->n_symbols = v16;
    if (fget_u8(f, &v8))
        return -1;
    p->tied_embed = v8;
    if (fget_be32(f, &p->seed))
        return -1;
    if (fget_f32(f, &p->init_range))
        return -1;
    if (fget_f32(f, &p->init_bias_v))
        return -1;
    if (fget_u8(f, &v8))
        return -1;
    p->more_relu = (v8 != 0);
    if (fget_sgd_opt(f, &p->sgd_opt))
        return -1;
    return 0;
}

/* XXX: we assume a single block is written */
static void my_write_func(void *opaque, const uint8_t *buf, size_t buf_len)
{
    FILE *fo = opaque;
    fput_be32(fo, buf_len);
    fwrite(buf, 1, buf_len, fo);
}

#define BLOCK_LEN 100000
#define STATS_AVG_COUNT (5000000 / BLOCK_LEN)

typedef struct {
    PutBitState pb;
} EncodeStream;

typedef uint16_t DataSymbol;

/* compress block_len bytes divided in n_streams */
void compress_block(TransformerModel *pred, FILE *fo,
                    const DataSymbol *block_buf, int block_len, float lr)
{
    EncodeStream *tab_st;
    int block_idx, c, c_in, stream_idx, cur_state, n_states, n_streams1, len;
    int n_states1, pos, block_stride, n_streams;
    int *input_syms, *output_syms;

    nc_sgd_opt_set_learning_rate(pred->sgd_opt, lr);
    
    n_streams = pred->n_streams;
    n_states = pred->n_states;
    block_stride = (block_len + n_streams - 1) / n_streams;

    input_syms = malloc(sizeof(input_syms[0]) * n_states * n_streams);
    output_syms = malloc(sizeof(output_syms[0]) * n_states * n_streams);
    tab_st = malloc(sizeof(tab_st[0]) * n_streams);
    
    for(stream_idx = 0; stream_idx < n_streams; stream_idx++) {
        EncodeStream *st = &tab_st[stream_idx];
        put_bit_init(&st->pb, my_write_func, fo);
    }
    trf_reset(pred);
    
    n_streams1 = min_int(block_len, n_streams);
    for(stream_idx = 0; stream_idx < n_streams1; stream_idx++) {
        EncodeStream *st = &tab_st[stream_idx];

        /* first char of each block */
        c = block_buf[stream_idx * block_stride];
        write_sym(&st->pb, pred->prob_output + stream_idx * pred->prob_output_stride, pred->n_symbols, c);
    }
    block_idx = 1;

    /* normal batches */
    while (block_len >= (n_states * n_streams + block_idx * n_streams)) {
        prof_start(PROF_TOTAL);
        for(cur_state = 0; cur_state < n_states; cur_state++) {
            for(stream_idx = 0; stream_idx < n_streams; stream_idx++) {
                input_syms[cur_state * n_streams + stream_idx] =
                    block_buf[stream_idx * block_stride +
                              block_idx + cur_state - 1];
            }
        }
        trf_eval(pred, -1, input_syms);
        for(cur_state = 0; cur_state < n_states; cur_state++) {
            for(stream_idx = 0; stream_idx < n_streams; stream_idx++) {
                EncodeStream *st = &tab_st[stream_idx];
                
                c = block_buf[stream_idx * block_stride +
                              block_idx + cur_state];
                prof_start(PROF_WRITE_SYM);
                write_sym(&st->pb, pred->prob_output +
                          (cur_state * n_streams + stream_idx) *
                          pred->prob_output_stride, pred->n_symbols, c);
                prof_end(PROF_WRITE_SYM);
                output_syms[cur_state * n_streams + stream_idx] = c;
            }
        }
        trf_eval_gradient(pred, output_syms);
        trf_update(pred);
        block_idx += n_states;
        prof_end(PROF_TOTAL);
    }
    
    len = block_len - block_idx * n_streams;
    if (len > 0) {
        /* no training for the last batch */
        n_states1 = (len + n_streams - 1) / n_streams;
        for(cur_state = 0; cur_state < n_states; cur_state++) {
            for(stream_idx = 0; stream_idx < n_streams; stream_idx++) {
                if (cur_state < n_states1) {
                    pos = stream_idx * block_stride + block_idx + cur_state - 1;
                    if (pos < block_len)
                        c_in = block_buf[pos];
                    else
                        c_in = 0;
                } else {
                    c_in = 0;
                }
                input_syms[cur_state * n_streams + stream_idx] = c_in;
            }
        }
        trf_eval(pred, -1, input_syms);
        for(cur_state = 0; cur_state < n_states1; cur_state++) {
            for(stream_idx = 0; stream_idx < n_streams; stream_idx++) {
                EncodeStream *st = &tab_st[stream_idx];
                pos = stream_idx * block_stride + block_idx + cur_state;
                if (pos < block_len) {
                    c = block_buf[pos];
                    write_sym(&st->pb, pred->prob_output +
                              (cur_state * n_streams + stream_idx) *
                              pred->prob_output_stride, pred->n_symbols, c);
                }
            }
        }
    }
    
    /* file output */
    fput_be32(fo, block_len);
    fput_f32(fo, lr);
    for(stream_idx = 0; stream_idx < n_streams; stream_idx++) {
        EncodeStream *st = &tab_st[stream_idx];
        put_bit_end(&st->pb);
    }
    fflush(fo);
    free(tab_st);
    free(input_syms);
    free(output_syms);
}

static const char *load_coefs_filename;
static BOOL quiet_flag;

static float compute_lr(const TransformerModelParams *p, int64_t pos)
{
    int i;
    float lr, t;
    int64_t pos0;

    pos0 = 0;
    for(i = 0; i < p->lr_steps; i++) {
        if (pos < p->lr_pos[i]) {
            if (i == 0)
                pos0 = 0;
            else
                pos0 = p->lr_pos[i - 1];
            t = (float)(pos - pos0) / (float)(p->lr_pos[i] - pos0);
            lr = p->lr[i] + t * (p->lr[i + 1] - p->lr[i]);
            return lr;
        }
    }
    return p->lr[p->lr_steps];
}

void compress_file(const TransformerModelParams *p,
                   const char *in_filename, const char *out_filename,
                   const char *debug_dir)
{
    FILE *f, *fo;
    TransformerModel *pred;
    Stats st_s, *st = &st_s;
    DataSymbol *block_buf;
    int block_len, n_streams, n_states, block_len_max, symb_shift, c;
    int64_t n_output_bytes;
    float lr;
    char out_filename_buf[1024];
    
    pred = trf_init(p);
    if (!quiet_flag) {
        trf_dump_params(stdout, pred, p, TRUE);
    }

    if (load_coefs_filename)
        nc_load_coefs(pred->model, load_coefs_filename);
    
    f = fopen(in_filename, "rb");
    if (!f) {
        perror(in_filename);
        exit(1);
    }

    n_streams = pred->n_streams;
    n_states = pred->n_states;
    
    block_len_max = (((BLOCK_LEN - 1) / n_states) * n_states + 1) * n_streams;
    block_buf = nc_malloc(block_len_max * sizeof(block_buf[0]));
    
    stats_init(st, debug_dir, "trfcp", in_filename, STATS_AVG_COUNT);
    if (st->fp) {
        fprintf(st->fp, "# ");
        trf_dump_params(st->fp, pred, p, TRUE);
        fflush(st->fp);
    }

    if (!out_filename) {
        snprintf(out_filename_buf, sizeof(out_filename_buf), "%s/%s",
                 st->debug_dir, "out.bin");
        out_filename = out_filename_buf;
    }
    
    fo = fopen(out_filename, "wb");
    if (!fo) {
        perror(out_filename);
        exit(1);
    }

    trf_write_file_header(fo, p);
    
    if (p->n_symbols <= 256)
        symb_shift = 0;
    else
        symb_shift = 1;

    for(;;) {
        for(block_len = 0; block_len < block_len_max; block_len++) {
            if (symb_shift == 0) {
                c = fgetc(f);
                if (c < 0)
                    break;
            } else {
                uint16_t v16;
                if (fget_be16(f, &v16))
                    break;
                c = v16;
            }
            if (c >= p->n_symbols) {
                fprintf(stderr, "Invalid symbol: %d\n", c);
                exit(1);
            }
            block_buf[block_len] = c;
        }
        if (block_len == 0)
            break;
        lr = compute_lr(p, st->n_input_bytes);
        compress_block(pred, fo, block_buf, block_len, lr);
        st->n_input_bytes += block_len;
        n_output_bytes = ftell(fo);
        st->learning_rate = lr;
        stats_dump(st, pred->model, n_output_bytes * 8);
    }
    nc_free(block_buf);
    
    n_output_bytes = ftell(fo);
    fclose(fo);
    fclose(f);

    stats_end(st, pred->model, n_output_bytes);
    
    trf_end(pred);

    nc_prof_dump();
}

typedef struct {
    uint8_t *buf;
    GetBitState gb;
} DecodeStream;

void decompress_file(const char *in_filename, const char *out_filename)
{
    TransformerModelParams p_s, *p = &p_s;
    FILE *f, *fo;
    TransformerModel *pred;
    int64_t byte_pos, ti;
    uint32_t block_len, cblock_len;
    int c, n_streams, n_states, block_stride, stream_idx, n_states1, n_streams1;
    int pos, len, block_idx, cur_state, c_in, symb_shift, i;
    int *input_syms, *output_syms;
    DataSymbol *block_buf;
    DecodeStream *dec_st;
    float lr;
    
    f = fopen(in_filename, "rb");
    if (!f) {
        perror(in_filename);
        exit(1);
    }

    if (trf_read_file_header(f, p) < 0) {
        fprintf(stderr, "Invalid file header\n");
        exit(1);
    }
    p->is_decoder = TRUE;
    
    fo = fopen(out_filename, "wb");
    if (!fo) {
        perror(out_filename);
        exit(1);
    }

    pred = trf_init(p);
    if (!quiet_flag) {
        trf_dump_params(stdout, pred, p, FALSE);
    }

    if (load_coefs_filename)
        nc_load_coefs(pred->model, load_coefs_filename);

    if (pred->n_symbols <= 256)
        symb_shift = 0;
    else
        symb_shift = 1;
    n_streams = pred->n_streams;
    n_states = pred->n_states;
    input_syms = malloc(sizeof(input_syms[0]) * n_states * n_streams);
    output_syms = malloc(sizeof(output_syms[0]) * n_states * n_streams);
    dec_st = nc_mallocz(sizeof(dec_st[0]) * n_streams);
    byte_pos = 0;
    
    ti = get_time_ms();
    for(;;) {
        if (fget_be32(f, &block_len) < 0)
            break;
        if (block_len == 0)
            break;
        if (fget_f32(f, &lr) < 0) {
            fprintf(stderr, "Could not read LR\n");
            exit(1);
        }
        block_stride = (block_len + n_streams - 1) / n_streams;

        nc_sgd_opt_set_learning_rate(pred->sgd_opt, lr);
        
        block_buf = nc_malloc(block_stride * n_streams * sizeof(block_buf[0]));
        for(stream_idx = 0; stream_idx < n_streams; stream_idx++) {
            DecodeStream *st = &dec_st[stream_idx];
            if (fget_be32(f, &cblock_len) < 0) {
                fprintf(stderr, "Could not read cblock_len\n");
                exit(1);
            }
            st->buf = nc_malloc(cblock_len);
            if (fread(st->buf, 1, cblock_len, f) != cblock_len) {
                fprintf(stderr, "Could not read input stream\n");
                exit(1);
            }
            get_bit_init(&st->gb, st->buf, cblock_len, NULL, NULL);
        }
        trf_reset(pred);


        /* first byte */
        n_streams1 = min_int(block_len, n_streams);
        for(stream_idx = 0; stream_idx < n_streams1; stream_idx++) {
            DecodeStream *st = &dec_st[stream_idx];
            c = read_sym(&st->gb, pred->prob_output + stream_idx * pred->prob_output_stride, pred->n_symbols);
            block_buf[stream_idx * block_stride] = c;
        }
        
        block_idx = 1;
        /* normal batches */
        while (block_len >= (n_states * n_streams + block_idx * n_streams)) {
            prof_start(PROF_TOTAL);
            for(cur_state = 0; cur_state < n_states; cur_state++) {
                for(stream_idx = 0; stream_idx < n_streams; stream_idx++) {
                    input_syms[cur_state * n_streams + stream_idx] = 0;
                }
            }

            for(cur_state = 0; cur_state < n_states; cur_state++) {
                for(stream_idx = 0; stream_idx < n_streams; stream_idx++) {
                    input_syms[cur_state * n_streams + stream_idx] = 
                        block_buf[stream_idx * block_stride +
                                  block_idx + cur_state - 1];
                }
                trf_eval(pred, cur_state, input_syms);
                for(stream_idx = 0; stream_idx < n_streams; stream_idx++) {
                    DecodeStream *st = &dec_st[stream_idx];
                    c = read_sym(&st->gb, pred->prob_output +
                                 (cur_state * n_streams + stream_idx) *
                                 pred->prob_output_stride, pred->n_symbols);
                    //                    printf("%d.%d: c=%d\n", cur_state, stream_idx, c);
                    block_buf[stream_idx * block_stride +
                              block_idx + cur_state] = c;
                    output_syms[cur_state * n_streams + stream_idx] = c;
                }
            }
            trf_eval_gradient(pred, output_syms);
            trf_update(pred);
            block_idx += n_states;
            prof_end(PROF_TOTAL);
        }
            
        /* last partial batch */
        len = block_len - block_idx * n_streams;
        if (len > 0) {
            /* no training for the last batch */
            n_states1 = (len + n_streams - 1) / n_streams;
            for(cur_state = 0; cur_state < n_states; cur_state++) {
                for(stream_idx = 0; stream_idx < n_streams; stream_idx++) {
                    input_syms[cur_state * n_streams + stream_idx] = 0;
                }
            }
            for(cur_state = 0; cur_state < n_states1; cur_state++) {
                for(stream_idx = 0; stream_idx < n_streams; stream_idx++) {
                    pos = stream_idx * block_stride + block_idx + cur_state - 1;
                    if (pos < block_len)
                        c_in = block_buf[pos];
                    else
                        c_in = 0;
                    input_syms[cur_state * n_streams + stream_idx] = c_in;
                }
                trf_eval(pred, cur_state, input_syms);

                for(stream_idx = 0; stream_idx < n_streams; stream_idx++) {
                    DecodeStream *st = &dec_st[stream_idx];
                    pos = stream_idx * block_stride + block_idx + cur_state;
                    if (pos < block_len) {
                        c = read_sym(&st->gb, pred->prob_output +
                                     (cur_state * n_streams + stream_idx) *
                                     pred->prob_output_stride, pred->n_symbols);
                        block_buf[pos] = c;
                    }
                }
            }
        }
        
        for(i = 0; i < block_len; i++) {
            if (symb_shift == 0)
                fputc(block_buf[i], fo);
            else
                fput_be16(fo, block_buf[i]);
        }
        nc_free(block_buf);

        for(stream_idx = 0; stream_idx < n_streams; stream_idx++) {
            DecodeStream *st = &dec_st[stream_idx];
            nc_free(st->buf);
        }

        byte_pos += block_len;
        if (!quiet_flag) {
            printf("%12" PRIu64 "\r", byte_pos);
            fflush(stdout);
        }
    }
    ti = get_time_ms() - ti;

    if (!quiet_flag) {
        printf("%12" PRIu64 " (%0.2f kS/s)\n", byte_pos, (double)byte_pos / ti);
    }
    
    fclose(fo);
    fclose(f);
    nc_free(dec_st);
    free(input_syms);
    free(output_syms);
    trf_end(pred);
    nc_prof_dump();
}

static struct option options[] = {
    { "coefs", required_argument },
    { "plot", required_argument },
    { "load_coefs", required_argument },

    { "n_layer", required_argument },
    { "d_model", required_argument },
    { "mem_len", required_argument },
    { "train_len", required_argument },
    { "d_pos", required_argument },
    { "init_range", required_argument },
    { "lr", required_argument },
    { "d_inner", required_argument },
    { "more_relu", required_argument },
    { "n_symb", required_argument },
    { "tied_embed", required_argument },
    { NULL },
};

static void help(void)
{
    printf("TRFCP version " CONFIG_VERSION", Copyright (c) 2018-2019 Fabrice Bellard\n"
           "Lossless data compression with Transformer\n"
           "usage: trfcp [options] c infile outfile\n"
           "                       d infile outfile\n"
           "\n"
           "General options:\n"
           "-h                            show this help\n"
           "-T n                          number of threads (default=1)\n"
           "-d path                       save logs & coefficients\n"
           "-q                            less verbose output\n"
           "\n"
           "Model options (only for compression):\n"
           "-n_layer n                    number of layers\n"
           "-d_model n\n"
           "-d_inner n\n"
           "-d_pos n\n"
           "-train_len n                  training segment length\n"
           "-mem_len n                    additional memory length \n"
           "-init_range n\n"
           "-lr lr0[,step0,lr1]...        learning rate schedule\n"
           "\n"
           "Debug options:\n"
           "-load_coefs file     load coefficients from file\n"
           );
    exit(1);
}

int main(int argc, char **argv)
{
    int c, option_index;
    TransformerModelParams p_s, *p = &p_s;
    const char *debug_dir;

    trf_set_default_params(p);
    debug_dir = NULL;
    for(;;) {
        c = getopt_long_only(argc, argv, "hqT:d:", options, &option_index);
        if (c == -1)
            break;
        switch(c) {
        case 0:
            switch(option_index) {
            case 2: /* load_coefs */
                load_coefs_filename = optarg;
                break;

            case 3:
                p->n_layer = atoi(optarg);
                break;
            case 4:
                p->d_model = atoi(optarg);
                p->d_key = p->d_model / p->n_head;
                p->d_value = p->d_key;
                break;
            case 5:
                p->mem_len = atoi(optarg);
                break;
            case 6:
                p->train_len = atoi(optarg);
                break;
            case 7:
                p->d_pos = atoi(optarg);
                break;
            case 8:
                p->init_range = strtod(optarg, NULL);
                break;
            case 9:
                {
                    const char *r = optarg;
                    p->lr_steps = 0;
                    p->lr[0] = strtod(r, (char **)&r);
                    for(;;) {
                        if (*r == '\0')
                            break;
                        if (*r != ',') {
                            fprintf(stderr, "expecting ','\n");
                            exit(1);
                        }
                        r++;
                        if (p->lr_steps >= LR_MAX_STEPS) {
                            fprintf(stderr, "Too many LR steps\n");
                            exit(1);
                        }
                        p->lr_pos[p->lr_steps] = lrint(strtod(r, (char **)&r));
                        if (*r != ',') {
                            fprintf(stderr, "expecting ','\n");
                            exit(1);
                        }
                        r++;
                        p->lr[++p->lr_steps] = strtod(r, (char **)&r);
                    }
                }
                break;
            case 10:
                p->d_inner = atoi(optarg);
                break;
            case 11:
                p->more_relu = (atoi(optarg) != 0);
                break;
            case 12:
                p->n_symbols = atoi(optarg);
                if (p->n_symbols < 2 || p->n_symbols > 65536) {
                    fprintf(stderr, "invalid number of symbols\n");
                    exit(1);
                }
                break;
            case 13:
                p->tied_embed = atoi(optarg);
                break;
            default:
                fprintf(stderr, "unknown option index: %d\n", option_index);
                exit(1);
            }
            break;
        case 'h':
            help();
            break;
        case 'q':
            quiet_flag = TRUE;
            break;
        case 'T':
            nb_threads = atoi(optarg);
            break;
        case 'd':
            debug_dir = optarg;
            break;
        default:
            exit(1);
        }
    }
            
    if (optind >= argc) 
        help();

    if (argv[optind][0] == 'c') {
        const char *out_filename;
        if (optind + 1 >= argc) 
            help();
        if ((optind + 2) < argc) {
            out_filename = argv[optind + 2];
        } else {
            out_filename = NULL;
            if (!debug_dir)
                help();
        }
        compress_file(p, argv[optind + 1], out_filename, debug_dir);
    } else if (argv[optind][0] == 'd') {
        if (optind + 2 >= argc) 
            help();
        decompress_file(argv[optind + 1], argv[optind + 2]);
    } else
#if 0
    if (argv[optind][0] == 't') {
        float eps;
        const char *var_name;
        var_name = argv[optind + 1];
        eps = 0.1;
        if ((optind + 2) < argc)
            eps = strtof(argv[optind + 2], NULL);
        gradient_test(var_name, eps);
    } else
#endif
    {
        help();
    }
    return 0;
}
